import os


from pymatgen.core import Structure

from maggma.builder import Builder

from emmet.common.utils import load_settings
__author__ = "Shyam Dwaraknath <shyamd@lbl.gov>"


module_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))
default_substrate_settings = os.path.join(
    module_dir, "settings", "electrodes.yaml")


class ElectrodeBuilder(Builder):

    def __init__(self, materials, electrodes, query=None, **kwargs):
        """
        Builds a convenient database of electrodes and properties

        Args:
            materials (Store): Store of materials documents
            electrodes (Store): Store of electrodes
            query (dict): dictionary to limit materials to be analyzed
        """

        self.materials = materials
        self.electrodes = electrodes
        self.query = query if query else {}
        self.working_ions = ["Li","Na","K","Rb","Cs","Mg","Cs"]
        self.redox_els = ["Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", "Zr", "Nb",
                          "Mo","Sn", "Sb", "W", "Re", "Bi"]

        super().__init__(sources=[materials],
                         targets=[electrodes],
                         **kwargs)

    def get_items(self):
        """
        Gets sets of entries from chemical systems that need to be processed 
        for possible electrodes

        Returns:
            generator of relevant entries from one chemical system
        """
        self.logger.info("Electrode Builder Started")

        # All updated chemical systems that contain at least one redox active elements
        q = dict(self.query)
        q.update(self.materials.lu_filter(self.thermo))
        q.update({"chemsys": {"$in": self.redox_els}})

        comps = self.materials.distinct("chemsys", q)

        self.logger.info(
            "Found {} compositions with new/updated materials".format(len(comps)))

        # Only yields maximal super sets: e.g. if ["A","B"] and ["A"] are both in the list, will only yield ["A","B"]
        # as this will calculate thermo props for all ["A"] compounds
        processed = set()

        # Start with the largest set to ensure we don"t miss superset/subset
        # relations
        for chemsys in sorted(comps, key=lambda x: len(x.split("-")), reverse=True):
            if chemsys not in processed:
                processed |= chemsys_permutations(chemsys)
                yield self.get_entries(chemsys)

    def get_electrode_entries(self, chemsys,working_ions):
        """
        Get all entries in a chemsys from materials

        Args:
            chemsys(str): a chemical system represented by string elements seperated by a dash (-)

        Returns:
            set(ComputedEntry): a set of entries for this system
        """

        self.logger.info("Getting entries for: {}".format(chemsys))

        new_q = dict(self.query)
        new_q["chemsys"] = {"$in": list(chemsys_permutations(chemsys))}
        fields = ["structure",  self.materials.key, "thermo.energy",
                  "unit_cell_formula", "calc_settings.is_hubbard",
                  "calc_settings.hubbards", "calc_settings.potcar_spec",
                  "calc_settings.run_type"]
        data = list(self.materials.query(fields, new_q))

        all_entries = []

        for d in data:
            parameters = {"is_hubbard": d["calc_settings"]["is_hubbard"],
                          "hubbards": d["calc_settings"]["hubbards"],
                          "potcar_spec": d["calc_settings"]["potcar_spec"],
                          "run_type": d["calc_settings"]["run_type"]
                          }

            entry = ComputedEntry(Composition(d["unit_cell_formula"]),
                                  d["thermo"]["energy"], 0.0, parameters=parameters,
                                  entry_id=d[self.materials.key],
                                  data={"oxide_type": oxide_type(Structure.from_dict(d["structure"]))})

            all_entries.append(entry)

        self.logger.info("Total entries in {} : {}".format(
            chemsys, len(all_entries)))

        return all_entries


def chemsys_permutations(chemsys):
    # Fancy way of getting every unique permutation of elements for all
    # possible number of elements:
    elements = chemsys.split("-")
    return {"-".join(sorted(c)) for c in
            chain(*[combinations(elements, i) for i in range(1, len(elements) + 1)])}




import numpy as np

from pymatgen.analysis.graphs import StructureGraph
from maggma.builder import Builder
from maggma.validator import StandardValidator

__author__ = "Matthew Horton <mkhorton@lbl.gov>"

class TopologySummaryValidator(StandardValidator):

    @property
    def schema(self):
        return {
            "type": "object",
            "properties":
                {
                    "task_id": {"type": "string"},
                    "methods": {"type": "array", "items": {"type": "string"}},
                    "distances": {"type": "array", "items": {"type": "array",
                                                             "items": {"type": "number"}}},
                    "preferred_graph": {"type": "object"},
                    "preferred_graph_method": {"type": "string"}
                },
            "required": ["task_id", "methods", "distances",
                         "preferred_graph", "preferred_graph_method"]
        }

    @property
    def msonable_keypaths(self):
        return {"graph": StructureGraph}


class TopologySummaryBuilder(Builder):
    def __init__(self, topology, topology_summary,
                 preferred_methods=('critic2_chgcar', 'MinimumOKeeffeNN'),
                 query=None, **kwargs):
        """
        Summarizes topology information (this is the 'materials' to
        topology's 'tasks').

        Args:
            topology (Store): Store of topology documents
            topology_summary (Store): Store of topology summary documents
            preferred_methods (Tuple): List of which topology methods to favor
            query (dict): dictionary to limit materials to be analyzed
        """

        self.topology = topology
        self.topology_summary = topology_summary
        self.preferred_methods = preferred_methods
        self.query = query if query else {}

        self.topology_summary.validator = TopologySummaryValidator()

        super().__init__(sources=[topology],
                         targets=[topology_summary],
                         **kwargs)

    def get_items(self):
        """
        Gets all materials that need topology analysis
        """

        self.logger.info("Topology Summary Builder Started")

        # Retrieve new task ids that were not included in previous summary
        new_task_ids_query = dict(self.query)
        new_task_ids_query.update(self.topology.lu_filter(self.topology_summary))

        new_task_ids = set([doc['task_id'] for doc in
                            self.topology.query(criteria=new_task_ids_query,
                                                properties=["task_id"])])

        self.logger.info("Found {} new topology docs to summarize.".format(len(new_task_ids)))

        # ensure we only aggregate a consistent set of user settings,
        # currently this is just default settings
        standard_critic2_user_settings = None

        for task_id in new_task_ids:

            topology_docs = list(self.topology.query(
                criteria={'task_id': task_id,
                          'successful': True,
                          '$or': [{'critic2_settings': {'$exists': False}},
                                  {'critic2_settings': standard_critic2_user_settings}]},
                properties=['task_id', 'method', 'graph', 'critic2_settings']))

            print(topology_docs)

            yield {
                'task_id': task_id,
                'topology_docs': topology_docs
            }

    def process_item(self, item):
        """
        Summarizes a set of StructureGraphs (calculates differences
        between them, picks one to be preferred output)
        """

        task_id = item['task_id']
        topology_docs = item['topology_docs']

        topology_summary = {'task_id': task_id}

        available_methods = [doc['method'] for doc in topology_docs]

        standard_methods = ('MinimumDistanceNN', 'MinimumOKeeffeNN', 'JMolNN',
                            'MinimumVIRENN', 'VoronoiNN', 'critic2_promol', 'critic2')

        sg = [StructureGraph.from_dict(doc['graph']) for doc in topology_docs]

        # standardized matrix to store graph distances
        distances = np.empty((len(standard_methods), len(standard_methods)))
        distances[:] = np.NaN

        # calculate distances between graphs
        # gives a measure of how 'different' one method
        # of determining bonding is from another method
        for i, method_a in enumerate(standard_methods):
            if method_a in available_methods:
                for j, method_b in enumerate(standard_methods[:i]):
                    if method_b in available_methods:

                        a_idx = available_methods.index(method_a)
                        b_idx = available_methods.index(method_b)

                        dist = sg[a_idx].diff(sg[b_idx])['dist']

                        distances[i][j] = dist
                        distances[j][i] = dist

        topology_summary['methods'] = standard_methods
        topology_summary['distances'] = distances

        # we want to store one canonical set of bonds for a material
        # we see if our preferred method is here
        preferred_method = None
        for method in reversed(self.preferred_methods):
            if method in available_methods:
                preferred_method = method

        if preferred_method:
            preferred_graph = sg[standard_methods.index(preferred_method)]
            topology_summary['preferred_graph'] = preferred_graph.as_dict()
            topology_summary['preferred_graph_method'] = preferred_method
            if 'ce_info' in preferred_graph.structure.site_properties:
                # for easier querying in mongo
                topology_summary['ce_info'] = preferred_graph.structure.site_properties['ce_info']

        self.logger.debug("Summarized bonding for {}".format(task_id))

        return topology_summary

    def update_targets(self, items):

        self.logger.info("Updating topology summary documents")
        self.topology_summary.update(items, key='task_id')



from pymatgen import Structure
from pymatgen.analysis.magnetism import CollinearMagneticStructureAnalyzer

from maggma.builder import Builder

__author__ = "Shyam Dwaraknath <shyamd@lbl.gov>"


class MagneticBuilder(Builder):
    def __init__(self, materials, magnetism, query={}, **kwargs):
        """
        Creates a magnetism collection for materials

        Args:
            materials (Store): Store of materials documents to match to
            magnetism (Store): Store of magnetism properties
            query (dict): dictionary to limit materials to be analyzed
        """

        self.materials = materials
        self.magnetism = magnetism
        self.query = query

        super().__init__(sources=[materials],
                         targets=[magnetism],
                         **kwargs)

    def get_items(self):
        """
        Gets all items to process into magnetismdocuments

        Returns:
            generator or list relevant tasks and materials to process into magnetism documents
        """
        self.logger.info("Magnestism Builder Started")



        # All relevant materials that have been updated since magnetism props
        # were last calculated
        q = dict(self.query)
        q.update(self.materials.lu_filter(self.magnetism))
        mats = list(self.materials.distinct(self.materials.key, q))
        self.logger.info(
            "Found {} new materials for magnetism data".format(len(mats)))
        for m in mats:
            yield self.materials.query(properties=[self.materials.key, "structure"],criteria={self.materials.key: m}).limit(1)[0]

    def process_item(self, item):
        """
        Process the tasks and materials into a dielectrics collection

        Args:
            item dict: a dict of material_id, structure, and tasks

        Returns:
            dict: a dieletrics dictionary  
        """

        struc = Structure.from_dict(item["structure"])
        msa = CollinearMagneticStructureAnalyzer(struc)

        magnetism = {
            self.magnetism.key : item[self.materials.key],
            "magnetism": {
                'ordering': msa.ordering.value
                }
        }

        return magnetism

    def update_targets(self, items):
        """
        Inserts the new task_types into the task_types collection

        Args:
            items ([([dict],[int])]): A list of tuples of materials to update and the corresponding processed task_ids
        """

        items = list(filter(None, items))

        if len(items) > 0:
            self.logger.info("Updating {} magnetism docs".format(len(items)))
            self.magnetism.update(docs=items)
        else:
            self.logger.info("No items to update")





from datetime import datetime
from itertools import groupby

import numpy as np
from scipy.interpolate import interp1d

from maggma.builder import Builder
from pydash import py_
from pymatgen import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from tqdm import tqdm


class XASAverager(Builder):
    def get_items(self):
        self.logger.info("Getting unprocessed mpids...")
        mpids = unprocessed_mpids(self.sources, self.targets)
        xas = self.sources[0]
        self.dt_fetch = datetime.utcnow()
        self.logger.info("Yielding XAS data for processing...")
        self.logger.info("{} unprocessed mpids".format(len(mpids)))
        for mp_id in tqdm(mpids):
            criteria = xas_criteria_base()
            criteria.update({'mp_id': mp_id})
            if not xas.query(criteria=criteria).count():
                raise Exception("No source docs for {}".format(mp_id))
            yield list(xas.query(criteria=criteria))

    def process_item(self, item):
        xas_docs_for_mpid = item
        mp_id = xas_docs_for_mpid[0]["mp_id"]
        if data_missing(xas_docs_for_mpid):
            return mark_invalid({"mp_id": mp_id})

        spectra = species_spectra(xas_docs_for_mpid)
        valids = [{
            "mp_id": mp_id,
            "element": element,
            "spectrum": [ary.tolist() for ary in spectrum]
        } for element, spectrum in spectra.items()]
        return [mark_valid(v) for v in valids]

    def update_targets(self, items):
        xas_averaged = self.targets[0]
        xas_averaged.ensure_index([("valid", 1), ("mp_id", 1)])
        xas_averaged.ensure_index([("mp_id", 1), ("element", 1)])
        xas_averaged.ensure_index([("chemsys", 1), ("element", 1)])
        valids, invalids = py_.partition(
            mark_lu(py_.flatten(items), xas_averaged.lu_field, self.dt_fetch),
            'valid')
        # Remove documents flagging now-valid data as invalid.
        xas_averaged.collection.delete_many(
            mark_invalid({
                "mp_id": {
                    "$in": py_.pluck(valids, 'mp_id')
                }
            }))
        bulk = xas_averaged.collection.initialize_ordered_bulk_op()
        for doc in valids:
            (bulk.find(py_.pick(doc, 'mp_id', 'element'))
                .upsert().replace_one(doc))
        for doc in invalids:
            (bulk.find(mark_invalid(py_.pick(doc, 'mp_id')))
                .upsert().replace_one(doc))
        bulk.execute()


def xas_criteria_base():
    return {'spectrum_type': 'XANES', 'edge': 'K'}


def unprocessed_mpids(sources, targets):
    xas = sources[0]
    materials = sources[1]
    xas_averaged = targets[0]
    mpids_marked_invalid = set(invalid_pks(xas_averaged, 'mp_id'))
    mpids_source_updated = set(
        updated_pks(xas, targets, 'mp_id', dt_map=lambda dt: dt.isoformat()))
    mpids_xas = xas.distinct('mp_id', criteria=xas_criteria_base())
    should = list(materials.query(
        criteria={'task_id': {'$in': mpids_xas}},
        properties={'_id': 0, 'task_id': 1, 'nelements': 1}))
    should = {(s['task_id'], s['nelements']) for s in should}
    actual = xas_averaged.collection.aggregate([
        {'$group': {'_id': '$mp_id', 'n': {'$sum': 1}}}
    ])
    actual = {(a['_id'], a['n']) for a in actual}
    mpids_build_incomplete = {s[0] for s in should - actual}
    return mpids_source_updated | (
        mpids_build_incomplete - mpids_marked_invalid)


def data_missing(xas_docs_for_mpid):
    """
    Do some sites have no spectra recorded?

    Checks symmetrically equivalent sites.
    """
    structure = Structure.from_dict(xas_docs_for_mpid[0]['structure'])
    ss = SymmSites(structure)
    absorbing_atoms = set([d['absorbing_atom'] for d in xas_docs_for_mpid])
    some_sites_absent = any(
        len(set(ss.get_equivalent_site_indices(i)) & absorbing_atoms) == 0
        for i in range(structure.num_sites))
    some_spectra_empty = any(
        len(d['spectrum']) == 0 for d in xas_docs_for_mpid)
    return some_sites_absent or some_spectra_empty


class SymmSites:
    def __init__(self, structure):
        self.structure = structure
        sa = SpacegroupAnalyzer(self.structure)
        symm_data = sa.get_symmetry_dataset()
        # equivalency mapping for the structure
        # i'th site in the input structure equivalent to eq_atoms[i]'th site
        self.eq_atoms = symm_data["equivalent_atoms"]

    def get_equivalent_site_indices(self, i):
        """
        Site indices in the structure that are equivalent to the given site i.
        """
        rv = np.argwhere(self.eq_atoms == self.eq_atoms[i]).squeeze().tolist()
        if isinstance(rv, int):
            rv = [rv]
        return rv


def species_spectra(xas_docs):
    """
    Get an equivalent-site-weighted spectrum for each specie in a structure.

    Args:
        xas_docs (list): MongoDB docs for all XAS XANES K-edge spectra
            for a structure.

    Returns:
        dict: {specie: (x, y)} a plottable (x, y) pair
            for each specie in the structure.
    """
    if not xas_docs:
        return []

    def absorbing_atom_element(d):
        site = d['structure']['sites'][d['absorbing_atom']]
        return site['species'][0]['element']

    xas_docs = sorted(xas_docs, key=absorbing_atom_element)
    spectrum = {}
    for element, group in groupby(xas_docs, absorbing_atom_element):
        spectrum[element] = site_weighted_spectrum(list(group))
    return spectrum


def site_weighted_spectrum(xas_docs, num_samples=200):
    """
    Equivalent-site-weighted spectrum for a specie in a structure.

    Args:
        xas_docs (list): MongoDB docs for all XAS XANES K-edge spectra
            for a specie for a structure.
        num_samples (int): Number of samples for interpolation.
            Original data has 100 data points.

    Returns:
        tuple: a plottable (x, y) pair for the spectrum
    """
    maxes, mins = [], []
    fs = []
    multiplicities = []

    for doc in xas_docs:
        energies = [e[0] for e in doc["spectrum"]]
        # Checking the multiplicities of sites
        s = Structure.from_dict(doc['structure'])
        sa = SpacegroupAnalyzer(s)
        ss = sa.get_symmetrized_structure()
        multiplicity = len(ss.find_equivalent_sites(s[doc['absorbing_atom']]))
        multiplicities.append(multiplicity)

        # Getting axis limits for each spectrum for the sites corresponding to
        # K-edge is a bit tricky, because the x-axis data points don't align
        # among different spectra for the same structure. So, prepare for
        # interpolation within the intersection of x-axis ranges.
        maxes.append(doc['spectrum'][-1][0])
        mins.append(doc['spectrum'][0][0])
        d0 = np.array(doc['spectrum'])
        # use 3rd-order spline interpolation for mu (idx 3) vs energy (idx 0).
        f = interp1d(
            d0[:, 0], d0[:, 3], kind='cubic', bounds_error=False, fill_value=0)
        fs.append(f)

    x_axis = np.linspace(max(mins), min(maxes), num=num_samples)
    weighted_spectrum = np.zeros(num_samples)
    sum_multiplicities = sum(multiplicities)
    for i in range(len(multiplicities)):
        weighted_spectrum += (
            multiplicities[i] * fs[i](x_axis)) / sum_multiplicities
    return (x_axis, weighted_spectrum)

# TODO: Migrate below functionality to emmet (or maggma) core / utils.

def mark_valid(doc):
    doc.update({"valid": True})
    return doc


def mark_invalid(doc):
    doc.update({"valid": False})
    return doc


def invalid_pks(target, pk):
    """Fetch values of primary key `pk` marked as invalid in target."""
    cursor = target.query(criteria=mark_invalid({}),
                          properties={'_id': 0, pk: 1})
    return py_.pluck(cursor, pk)


def updated_pks(source, targets, pk, dt_map=None):
    """Fetch primary key values that have new source data."""
    lu_targets = max([t.last_updated for t in targets])
    lu_targets = dt_map(lu_targets) if dt_map else lu_targets
    lu_filter = {source.lu_field: {'$gt': lu_targets}}
    criteria = xas_criteria_base()
    criteria.update(lu_filter)
    return source.distinct(pk, criteria=criteria)


def mark_lu(docs, lu_key, lu_val):
    """Mark documents with last-updated info."""
    return [py_.assign(d, {lu_key: lu_val}) for d in docs]




from itertools import chain
from pymatgen import Structure
from pymatgen.entries.computed_entries import ComputedStructureEntry
from emmet.vasp.task_tagger import task_type
from maggma.builder import Builder
from pydash.objects import get

__author__ = "Shyam Dwaraknath <shyamd@lbl.gov>"


class MLStructuresBuilder(Builder):

    def __init__(self, tasks, ml_strucs, task_types= ("Structure Optimization",),query=None, **kwargs):
        """
        Creates a collection of structures, energies, forces, and stresses for machine learning efforts

        Args:
            tasks (Store): Store of task documents
            ml_strucs (Store): Store of materials documents to generate
            tasK_types (list): list of substrings for task_types to process
        """

        self.tasks = tasks
        self.ml_strucs = ml_strucs
        self.task_types = task_types
        self.query = query if query else None
        super().__init__(sources=[tasks],
                         targets=[ml_strucs],
                         **kwargs)

    def get_items(self):
        """
        Gets all items to process into materials documents

        Returns:
            generator or list relevant tasks and materials to process into materials documents
        """

        self.logger.info("Machine Learning Structure Database Builder Started")
        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Get all processed tasks:
        q = dict(self.query)
        q["state"] = "successful"
        q["calcs_reversed"] = {"$exists": 1}

        all_tasks = set(self.tasks.distinct("task_id", q))
        processed_tasks = set(self.ml_strucs.distinct("task_id"))
        to_process_tasks = all_tasks - processed_tasks

        self.logger.info(
            "Found {} tasks to extract information from".format(len(to_process_tasks)))

        for t_id in to_process_tasks:
            task = self.tasks.query_one(properties=["task_id","orig_inputs","calcs_reversed"],criteria={"task_id": t_id})
            yield task

    def process_item(self, task):
        """
        Process the tasks into a list of materials

        Args:
            task [dict] : a task doc

        Returns:
            list of C
        """

        t_type = task_type(get(task, 'orig_inputs'))
        entries = []

        if not any([t in t_type for t in self.task_types]):
            return []

        is_hubbard = get(task, "input.is_hubbard", False)
        hubbards = get(task, "input.hubbards", [])
        i = 0

        for calc in task.get("calcs_reversed", []):

            parameters = {"is_hubbard": is_hubbard,
                          "hubbards": hubbards,
                          "potcar_spec": get(calc, "input.potcar_spec", []),
                          "run_type": calc.get("run_type", "GGA")
                          }

            for step_num, step in enumerate(get(calc, "output.ionic_steps")):
                struc = Structure.from_dict(step.get("structure"))
                forces = calc.get("forces", [])
                if forces:
                    struc.add_site_property("forces", forces)
                stress = calc.get("stress", None)
                data = {"stress": stress} if stress else {}
                data["step"] =step_num
                c = ComputedStructureEntry(structure=struc,
                                           correction=0,
                                           energy=step.get("e_wo_entrp"),
                                           parameters=parameters,
                                           entry_id="{}-{}".format(task[self.tasks.key],i),
                                           data=data)
                i += 1

                d = c.as_dict()
                d["chemsys"] = '-'.join(
                    sorted(set([e.symbol for e in struc.composition.elements])))
                d["task_type"] = task_type(get(calc, 'input'))
                d["calc_name"] = get(calc, "task.name")
                d["task_id"] = task[self.tasks.key]

                entries.append(d)

        return entries

    def update_targets(self, items):
        """
        Inserts the new entires into the task_types collection

        Args:
            items ([([dict],[int])]): A list of tuples of materials to update and the corresponding processed task_ids
        """
        items = [i for i in filter(None, chain.from_iterable(items))]

        if len(items) > 0:
            self.logger.info("Updating {} entries".format(len(items)))
            self.ml_strucs.update(docs=items)
        else:
            self.logger.info("No items to update")

    def ensure_indexes(self):
        """
        Ensures indexes on the tasks and materials collections
        :return:
        """

        # Basic search index for tasks
        self.ml_strucs.ensure_index("entry_id")
        self.ml_strucs.ensure_index("chemsys")
        self.ml_strucs.ensure_index(self.ml_strucs.lu_field)



import logging
import os
from datetime import datetime

from monty.json import jsanitize
from monty.serialization import loadfn

from pymatgen.core.structure import Structure
from pymatgen.analysis.diffraction.xrd import XRDCalculator, WAVELENGTHS

from maggma.builder import Builder

__author__ = "Shyam Dwaraknath <shyamd@lbl.gov>"

module_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))
default_xrd_settings = os.path.join(
    module_dir, "settings", "xrd.json")


class DiffractionBuilder(Builder):

    def __init__(self, materials, diffraction, xrd_settings=None, query=None, **kwargs):
        """
        Calculates diffraction patterns for materials

        Args:
            materials (Store): Store of materials documents
            diffraction (Store): Store of diffraction data such as formation energy and decomposition pathway
            xrd_settings (Store): Store of xrd settings
            query (dict): dictionary to limit materials to be analyzed
        """

        self.materials = materials
        self.diffraction = diffraction
        self.xrd_settings = xrd_settings if xrd_settings else default_xrd_settings
        self.query = query if query else {}
        self.__settings = loadfn(self.xrd_settings)

        super().__init__(sources=[materials],
                         targets=[diffraction],
                         **kwargs)

    def get_items(self):
        """
        Gets all materials that need a new XRD 

        Returns:
            generator of materials to calculate xrd
        """

        self.logger.info("Diffraction Builder Started")

        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # All relevant materials that have been updated since diffraction props
        # were last calculated
        q = dict(self.query)
        q.update(self.materials.lu_filter(self.diffraction))
        mats = list(self.materials.distinct(self.materials.key, q))
        self.logger.info(
            "Found {} new materials for diffraction data".format(len(mats)))
        for m in mats:
            yield self.materials.query(properties=[self.materials.key, "structure"], criteria={self.materials.key: m}).limit(1)[0]

    def process_item(self, item):
        """
        Calculates diffraction patterns for the structures

        Args:
            item (dict): a dict with a material_id and a structure

        Returns:
            dict: a diffraction dict
        """
        self.logger.debug("Calculating diffraction for {}".format(
            item[self.materials.key]))

        struct = Structure.from_dict(item['structure'])

        xrd_doc = {"xrd": self.get_xrd_from_struct(struct)}
        xrd_doc[self.diffraction.key] = item[self.materials.key]

        return xrd_doc

    def get_xrd_from_struct(self, structure):
        doc = {}

        for xs in self.__settings:
            xrdcalc = XRDCalculator(wavelength="".join([xs['target'], xs['edge']]),
                                    symprec=xs.get('symprec', 0))

            pattern = jsanitize(xrdcalc.get_xrd_pattern(
                structure, two_theta_range=xs['two_theta']).as_dict())
            # TODO: Make sure this is what the website actually needs
            d = {'wavelength': {'element': xs['target'],
                                'in_angstroms': WAVELENGTHS["".join([xs['target'], xs['edge']])]},
                 'pattern': pattern}
            doc[xs['target']] = d
        return doc

    def update_targets(self, items):
        """
        Inserts the new task_types into the task_types collection

        Args:
            items ([[dict]]): a list of list of thermo dictionaries to update
        """
        items = list(filter(None, items))

        if len(items) > 0:
            self.logger.info("Updating {} diffraction docs".format(len(items)))
            self.diffraction.update(docs=items)
        else:
            self.logger.info("No items to update")

    def ensure_indexes(self):
        """
        Ensures indexes on the tasks and materials collections
        :return:
        """
        # Search index for materials
        self.materials.ensure_index(self.materials.key, unique=True)

        # Search index for materials
        self.diffraction.ensure_index(self.diffraction.key, unique=True)



import logging
import tempfile
import gridfs
import os

from abipy.dfpt.phonons import get_dyn_mat_eigenvec, match_eigenvectors
from abipy.dfpt.anaddbnc import AnaddbNcFile
from monty.json import jsanitize
from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine
from pymatgen.core.structure import Structure
from pymatgen.io.abinit.abiobjects import KSampling
from pymatgen.symmetry.bandstructure import HighSymmKpath
import numpy as np
from abipy.abio.inputs import AnaddbInput
from abipy.core.abinit_units import Ha_cmm1
from abipy.flowtk.tasks import AnaddbTask, TaskManager
from abipy.dfpt.ddb import AnaddbError
from abipy.core.abinit_units import eV_to_THz

from maggma.builder import Builder


#TODO - handle possible other sources for the anaddb netcdf files?
#     - add input more parameters to tune anaddb calculation?
#     - identify the warning for the materials? (e.g. large ASR and CNSR breaking)
#     - add a second collection with the phononwebsite format?
#     - store the anaddb netcdf output to speedup possible rerunning of the builder?
#     - store the PhononBandStructureSymmLine in gridfs

class PhononBuilder(Builder):
    def __init__(self, materials, phonon, query=None, **kwargs):
        """
        CCreates a phonon collection for materials

        Args:
            materials (Store): Store of materials documents
            phonon (Store): Store of diffraction data such as formation energy and decomposition pathway
            query (dict): dictionary to limit materials to be analyzed
        """

        self.materials = materials
        self.phonon = phonon

        if query is None:
            query = {}
        self.query = query

        super().__init__(sources=[materials],
                         targets=[phonon],
                         **kwargs)

    def get_items(self):
        """
        Gets all materials that need phonons

        Returns:
            generator of materials to calculate xrd
        """

        self.logger.info("Phonon Builder Started")

        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # All relevant materials that have been updated since diffraction props were last calculated
        q = dict(self.query)
        q.update(self.materials.lu_filter(self.phonon))
        mats = list(self.materials().find(q, {"mp_id": 1}))
        self.logger.info("Found {} new materials for phonon data".format(len(mats)))

        # list of properties queried from the results DB
        # basic informations
        projection = {"mp_id": 1, "spacegroup.number":1}
        # input data
        projection.update({"abinit_input.structure": 1, "abinit_input.ngkpt": 1, "abinit_input.shiftk": 1,
                           "abinit_input.ecut": 1, "abinit_input.ngqpt": 1})
        # file ids to be fetched
        projection["abinit_output.ddb_id"] = 1

        # initialize the gridfs
        ddbfs = gridfs.GridFS(self.materials().database, "ddb_fs")

        for m in mats:
            item = self.materials().find_one(m, projection)

            # download the DDB file and add the path to the item
            with tempfile.NamedTemporaryFile() as f:
                f.write(ddbfs.get(m["abinit_output.ddb_id"]).read())
                item["ddb_path"] = f.name

            yield item

    def process_item(self, item):
        """
        Generates the full phonon document from an item

        Args:
            item (dict): a dict extracted from the phonon calculations results and the path of the
                downloaded DDB file.

        Returns:
            dict: a diffraction dict
        """
        self.logger.debug("Processing phonon item for {}".format(item['mp_id']))

        try:

            structure = Structure.from_dict(item["abinit_input.structure"])

            ph_doc = {"structure": structure.as_dict()}
            ph_doc["phonon"] = self.get_phonon_properties(item)
            ph_doc[self.phonon.key] = item[self.materials.key]

            return ph_doc
        except Exception as e:
            self.logger.warning(
                "Error generating the phonon properties for {}: {}".format(item["mp_id"], e))
            return None

    def get_phonon_properties(self, item):
        """
        Extracts the phonon properties from the item
        """

        # the temp dir should still exist when using the objects as some readings are done lazily
        with tempfile.TemporaryDirectory() as workdir:
            phbst_file, phdos_file, ananc_file, labels_list = self.run_anaddb(item, workdir=workdir)

            phbands = phbst_file.phbands
            phbands.read_non_anal_from_file(phbst_file.filepath)

            symm_line_bands = self.get_pmg_bs(phbands, labels_list)

            complete_dos = phdos_file.to_pymatgen()
            phdos = phdos_file.phdos

            tstart, tstop, nt = 5, 800, 160
            temp = np.linspace(tstart, tstop, nt)

            thermo = {"temperature": temp.tolist(),
                      "entropy": phdos.get_entropy(tstart, tstop, nt).values.tolist(),
                      "cv": phdos.get_cv(tstart, tstop, nt).values.tolist(),
                      "free_energy": phdos.get_free_energy(tstart, tstop, nt).values.tolist(),
                      }

            data = {"dos": complete_dos.as_dict(),
                    "bs": symm_line_bands.as_dict(),
                    "thermodynamic": thermo,
                    "becs": ananc_file.becs.values.tolist()}

            return jsanitize(data)

    def run_anaddb(self, item, workdir):
        """
        Runs anaddb an reads the outputs
        """

        anaddb_inp, labels_list = self.get_anaddb_input(item)

        #TODO should this be an input?
        tm = TaskManager.from_user_config()

        ddb_path = item["ddb_path"]

        task = AnaddbTask.temp_shell_task(anaddb_inp, ddb_node=ddb_path, workdir=workdir, manager=tm)

        # Run the task here.
        task.start_and_wait(autoparal=False)

        report = task.get_event_report()
        if not report.run_completed:
            raise AnaddbError(task=task, report=report)

        phbst_file = task.open_phbst()
        phdos_file = task.open_phdos()
        ananc_file = AnaddbNcFile.from_file(os.path.join(workdir, "anaddb.nc"))

        return phbst_file, phdos_file, ananc_file, labels_list



    def get_anaddb_input(self, item):
        """
        creates the AnaddbInput object. It also returns the list of qpoints labels for generating the
        PhononBandStructureSymmLine.
        """

        ngqpt = item["abinit_input.ngqpt"]
        q1shft = [(0, 0, 0)]

        structure = Structure.from_dict(item["abinit_input.structure"])

        hs = HighSymmKpath(structure, symprec=1e-2)

        spgn = hs._sym.get_space_group_number()
        if spgn != item["spacegroup.number"]:
            raise RuntimeError("Parsed specegroup number {} does not match "
                               "calculation spacegroup {}".format(spgn, item["spacegroup.number"]))

        # for the moment use gaussian smearing
        prtdos = 1
        dossmear = 3 / Ha_cmm1
        lo_to_splitting = True
        dipdip = 1
        asr = 2
        chneut = 1
        ng2qppa = 50000

        inp = AnaddbInput(structure, comment="ANADB input for phonon bands and DOS")

        inp.set_vars(
            ifcflag=1,
            ngqpt=np.array(ngqpt),
            q1shft=q1shft,
            nqshft=len(q1shft),
            asr=asr,
            chneut=chneut,
            dipdip=dipdip,
        )

        # Parameters for the dos.
        ng2qpt = KSampling.automatic_density(structure, kppa=ng2qppa).kpts[0]
        inp.set_vars(prtdos=prtdos, dosdeltae=None, dossmear=dossmear, ng2qpt=ng2qpt)

        # Parameters for the BS
        qpts, labels_list = hs.get_kpoints(line_density=18, coords_are_cartesian=False)

        n_qpoints = len(qpts)
        qph1l = np.zeros((n_qpoints, 4))

        qph1l[:, :-1] = qpts
        qph1l[:, -1] = 1

        inp['qph1l'] = qph1l.tolist()
        inp['nph1l'] = n_qpoints


        if lo_to_splitting:
            kpath = hs.kpath
            directions = []
            for qptbounds in kpath['path']:
                for i, qpt in enumerate(qptbounds):
                    if np.array_equal(kpath['kpoints'][qpt], (0, 0, 0)):
                        # anaddb expects cartesian coordinates for the qph2l list
                        if i > 0:
                            directions.extend(structure.lattice.reciprocal_lattice_crystallographic.get_cartesian_coords(
                                kpath['kpoints'][qptbounds[i - 1]]))
                            directions.append(0)

                        if i < len(qptbounds) - 1:
                            directions.extend(structure.lattice.reciprocal_lattice_crystallographic.get_cartesian_coords(
                                kpath['kpoints'][qptbounds[i + 1]]))
                            directions.append(0)

            if directions:
                directions = np.reshape(directions, (-1, 4))
                inp.set_vars(
                    nph2l=len(directions),
                    qph2l=directions
                )

        return inp, labels_list

    def get_pmg_bs(self, phbands, labels_list):
        """
        Generates a PhononBandStructureSymmLine starting from a abipy PhononBands object

        Args:
            phbands: abipy PhononBands
            labels_list: list of labels used to generate the path
        Returns:
            An instance of PhononBandStructureSymmLine
        """

        structure = phbands.structure

        n_at = len(structure)

        qpts = np.array(phbands.qpoints.frac_coords)
        ph_freqs = np.array(phbands.phfreqs)
        displ = np.array(phbands.phdispl_cart)

        labels_dict = {}

        for i, (q, l) in enumerate(zip(qpts, labels_list)):
            if l:
                labels_dict[l] = q
                # set LO-TO at gamma
                if "Gamma" in l:
                    if i > 0 and not labels_list[i-1]:
                        ph_freqs[i] = phbands._get_non_anal_freqs(qpts[i-1])
                        displ[i] = phbands._get_non_anal_phdispl(qpts[i-1])
                    if i < len(qpts)-1 and not labels_list[i+1]:
                        ph_freqs[i] = phbands._get_non_anal_freqs(qpts[i+1])
                        displ[i] = phbands._get_non_anal_phdispl(qpts[i+1])

        ind = self.match_bands(displ, phbands.structure, phbands.amu)

        ph_freqs = ph_freqs[np.arange(len(ph_freqs))[:, None], ind]
        displ = displ[np.arange(displ.shape[0])[:, None, None],
                      ind[..., None],
                      np.arange(displ.shape[2])[None, None, :]]

        ph_freqs = np.transpose(ph_freqs) * eV_to_THz
        displ = np.transpose(np.reshape(displ, (len(qpts), 3*n_at, n_at, 3)), (1, 0, 2, 3))

        return PhononBandStructureSymmLine(qpoints=qpts, frequencies=ph_freqs,
                                           lattice=structure.reciprocal_lattice,
                                           has_nac=phbands.non_anal_ph is not None, eigendisplacements=displ,
                                           labels_dict=labels_dict, structure=structure)

    def match_bands(self, displ, structure, amu):
        """
        Match the phonon bands of neighboring q-points based on the scalar product of the eigenvectors
        """
        eigenvectors = get_dyn_mat_eigenvec(displ, structure, amu)
        ind = np.zeros(displ.shape[0:2], dtype=np.int)

        ind[0] = range(displ.shape[1])

        for i in range(1, len(eigenvectors)):
            match = match_eigenvectors(eigenvectors[i - 1], eigenvectors[i])
            ind[i] = [match[m] for m in ind[i - 1]]

        return ind

    def update_targets(self, items):
        """
        Inserts the new task_types into the task_types collection

        Args:
            items ([[dict]]): a list of list of phonon dictionaries to update
        """
        items = list(filter(None, items))

        if len(items) > 0:
            self.logger.info("Updating {} phonon docs".format(len(items)))
            self.phonon.update(docs=items)
        else:
            self.logger.info("No items to update")


    def ensure_indexes(self):
        """
        Ensures indexes on the tasks and materials collections
        :return:
        """
        # Search index for materials
        self.materials.ensure_index(self.materials.key, unique=True)

        # Search index for materials
        self.phonon.ensure_index(self.diffraction.key, unique=True)


from datetime import datetime
import os
import string
import traceback
from ast import literal_eval
from pymongo import ASCENDING, DESCENDING

from monty.serialization import loadfn
from monty.json import jsanitize

from maggma.builder import Builder
from pydash.objects import get, set_, has


# Import for crazy things this builder needs
from pymatgen.io.cif import CifWriter
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core.composition import Composition
from pymatgen import Structure
from pymatgen.analysis.structure_analyzer import oxide_type
from pymatgen.analysis.bond_valence import BVAnalyzer
from pymatgen.analysis.structure_analyzer import RelaxationAnalyzer
from pymatgen.analysis.diffraction.xrd import XRDPattern
from pymatgen.analysis.magnetism import CollinearMagneticStructureAnalyzer


__author__ = "Shyam Dwaraknath <shyamd@lbl.gov>"


module_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))
mp_conversion_dict = loadfn(os.path.join(
    module_dir, "settings", "mp_conversion.json"))


SANDBOXED_PROPERTIES = {"e_above_hull": "e_above_hull",
                        "decomposes_to": "decomposes_to"}

latt_para_interval = [1.50 - 1.96 * 3.14, 1.50 + 1.96 * 3.14]
vol_interval = [4.56 - 1.96 * 7.82, 4.56 + 1.96 * 7.82]


class MPBuilder(Builder):

    def __init__(self, materials,  mp_materials, thermo=None, electronic_structure=None, magnetism=None, snls=None, xrd=None, elasticity=None, piezo=None, icsd=None, query=None,  **kwargs):
        """
        Creates a MP Website style materials doc.
        This builder is a bit unweildy as MP will eventually move to a new format
        Written for backwards compatability with previous infrastructure

        Args:
            tasks (Store): Store of task documents
            materials (Store): Store of materials documents
            mp_web (Store): Store of the mp style website docs, This will also make an electronic_structure collection and an es_plot gridfs
        """
        self.materials = materials
        self.mp_materials = mp_materials
        self.electronic_structure = electronic_structure
        self.snls = snls
        self.thermo = thermo
        self.query = query if query else None
        self.icsd = icsd
        self.xrd = xrd
        self.elasticity = elasticity
        self.piezo = piezo
        self.magnetism = magnetism

        sources = list(filter(None, [
                       materials, thermo, electronic_structure, magnetism, snls, elasticity, piezo, icsd, xrd]))

        super().__init__(sources=sources,
                         targets=[mp_materials],
                         **kwargs)

    def get_items(self):
        """
        Gets all materials that need a new MP Style materials document

        Returns:
            generator of materials to calculate xrd
        """

        self.logger.info("MP Website Builder Started")

        # All relevant materials that have been updated since MP Website Materials
        # were last calculated
        q = dict(self.query)
        q.update(self.materials.lu_filter(self.mp_materials))
        mats = list(self.materials.distinct(self.materials.key, q))
        self.logger.info(
            "Found {} new materials for the website".format(len(mats)))

        for m in mats:

            doc = {"material": self.materials.query_one(
                criteria={self.materials.key: m})}

            if self.electronic_structure:
                doc["electronic_structure"] = self.electronic_structure.query_one(
                    criteria={self.electronic_structure.key: m, "band_gap": {"$exists": 1}})

            if self.elasticity:
                doc["elasticity"] = self.elasticity.query_one(
                    criteria={self.elasticity.key: m})

            if self.piezo:
                doc["piezo"] = self.piezo.query_one(
                    criteria={self.piezo.key: m})

            if self.thermo:
                doc["thermo"] = self.thermo.query_one(
                    criteria={self.thermo.key: m})

            if self.snls:
                doc["snl"] = self.snls.query_one(
                    criteria={self.snls.key: m})

            if self.icsd:
                doc["icsds"] = list(self.icsd.query(
                    criteria={"chemsys": doc["material"]["chemsys"]}))

            if self.magnetism:
                doc["magnetism"] = self.magnetism.query_one(
                    criteria={self.magnetism.key: m})

            if self.xrd:
                doc["xrd"] = self.xrd.query_one(
                    criteria={self.xrd.key: m})

            self.logger.debug("Working on {}".format(m))
            yield doc

    def process_item(self, item):

        new_mat = item["material"]

        mat = old_style_mat(new_mat)

        if item.get("xrd", None):
            xrd = item["xrd"]
            add_xrd(mat, xrd)

        if item.get("electronic_structure", None):
            es = item["electronic_structure"]
            add_es(mat, new_mat, es)

        if item.get("piezo", None):
            pass

        if item.get("elasticity", None):
            pass

        if item.get("thermo", None):
            thermo = item["thermo"]
            add_thermo(mat, thermo)

        if item.get("snl", None):
            snl = item["snl"]
            add_snl(mat, snl)

            if item.get("icsds", None):
                icsds = item["icsds"]
                if icsds:
                    add_icsd(mat, icsds)

        sandbox_props(mat)
        return jsanitize(mat)

    def update_targets(self, items):
        """
        Inserts the new task_types into the task_types collection

        Args:
            items ([[dict]]): a list of list of thermo dictionaries to update
        """
        items = list(filter(None, items))

        if len(items) > 0:
            self.logger.info(
                "Updating {} mp materials docs".format(len(items)))
            self.mp_materials.update(docs=items)
        else:
            self.logger.info("No items to update")

    def ensure_indexes(self):
        """
        Ensures indexes on the tasks and materials collections
        :return:
        """

        # Basic search index for tasks
        self.tasks.ensure_index("task_id", unique=True)
        self.tasks.ensure_index("state")
        self.tasks.ensure_index("formula_pretty")
        self.tasks.ensure_index(self.tasks.lu_field)

        # Search index for materials
        self.materials.ensure_index(self.materials.key, unique=True)
        self.materials.ensure_index("task_ids")

#
#
#
#
# THIS SECTION DEFINES EXTRA FUNCTIONS THAT MODIFY THE MAT DOC PER MP DOC STRUCTURE
#
#


def old_style_mat(new_mat):
    """
    Creates the base document for the old MP mapidoc style from the new document structure
    """

    mat = {}
    for mp, new_key in mp_conversion_dict.items():
        if has(new_mat, new_key):
            set_(mat, mp, get(new_mat, new_key))

    mat["is_orderd"] = True
    mat["is_compatible"] = True

    struc = Structure.from_dict(mat["structure"])
    mat["oxide_type"] = oxide_type(struc)
    mat["reduced_cell_formula"] = struc.composition.as_dict()
    mat["full_formula"] = "".join(struc.formula.split())
    vals = sorted(mat["reduced_cell_formula"].values())
    mat["anonymous_formula"] = {string.ascii_uppercase[i]: float(vals[i])
                                for i in range(len(vals))}

    set_(mat, "pseudo_potential.functional", "PBE")

    set_(mat, "pseudo_potential.labels", [p["titel"].split()[1] for p in get(new_mat, "calc_settings.potcar_spec")])
    mat["ntask_ids"] = len(get(new_mat, "task_ids"))
    set_(mat, "pseudo_potential.pot_type", "paw")
    add_bv_structure(mat)
    add_blessed_tasks(mat, new_mat)
    add_cifs(mat)

    return mat


def add_es(mat, new_mat, es):

    bs_origin = None
    dos_origin = None
    try:
        bs_origin = next((origin for origin in new_mat.get("origins", []) if "Line" in origin["task_type"]), None)
        dos_origin = next((origin for origin in new_mat.get("origins", []) if "Uniform" in origin["task_type"]), None)

        if bs_origin:
            u_type = "GGA+U" if "+U" in bs_origin["task_type"] else "GGA"
            set_(mat, "band_structure.{}.task_id".format(u_type), bs_origin["task_id"])
            if has(es, "band_gap"):
                set_(mat, "band_gap.search_gap", get(es, "band_gap"))

        if dos_origin:
            u_type = "GGA+U" if "+U" in dos_origin["task_type"] else "GGA"
            set_(mat, "dos.{}.task_id".format(u_type), dos_origin["task_id"])

    except Exception as e:
        print("Error in adding electronic structure: {}".format(e))

    mat["has_bandstructure"] = bool(bs_origin) and bool(dos_origin)


def add_blessed_tasks(mat, new_mat):
    blessed_tasks = {}
    for doc in new_mat["origins"]:
        blessed_tasks[doc["task_type"]] = doc["task_id"]

    mat["blessed_tasks"] = blessed_tasks


def add_bv_structure(doc):
    struc = Structure.from_dict(doc["structure"])
    try:
        bva = BVAnalyzer()
        bv_struct = bva.get_oxi_state_decorated_structure(struc)
        doc["bv_structure"] = bv_struct.as_dict()
    except Exception as e:
        print("BVAnalyzer error: {}".format(e))


def add_cifs(doc):
    struc = Structure.from_dict(doc["structure"])
    sym_finder = SpacegroupAnalyzer(struc, symprec=0.1)
    doc["cif"] = str(CifWriter(struc))
    doc["cifs"] = {}
    if sym_finder.get_hall():
        primitive = sym_finder.get_primitive_standard_structure()
        conventional = sym_finder.get_conventional_standard_structure()
        refined = sym_finder.get_refined_structure()
        doc["cifs"]["primitive"] = str(CifWriter(primitive))
        doc["cifs"]["refined"] = str(CifWriter(refined))
        doc["cifs"]["conventional_standard"] = str(CifWriter(conventional))
        doc["spacegroup"]["symbol"] = sym_finder.get_space_group_symbol()
        doc["spacegroup"]["number"] = sym_finder.get_space_group_number()
        doc["spacegroup"]["point_group"] = sym_finder.get_point_group_symbol()
        doc["spacegroup"]["crystal_system"] = sym_finder.get_crystal_system()
        doc["spacegroup"]["hall"] = sym_finder.get_hall()
    else:
        doc["cifs"]["primitive"] = None
        doc["cifs"]["refined"] = None
        doc["cifs"]["conventional_standard"] = None
        _log.error("Can't get alternative cells for task_id {i}".format(
            i=doc["task_id"]))


def add_xrd(mat, xrd):
    mat["xrd"] = {}
    for el, doc in xrd["xrd"].items():
        el_doc = {}
        el_doc["meta"] = ["amplitude", "hkl", "two_theta", "d_spacing"]
        el_doc["created_at"] = datetime.now().isoformat()
        el_doc["wavelength"] = doc["wavelength"]

        xrd_pattern = XRDPattern.from_dict(doc["pattern"])
        el_doc["pattern"] = [[float(intensity),
                              [int(x) for x in literal_eval(list(hkls.keys())[0])],
                              two_theta,
                              float(d_hkl)] for two_theta, intensity, hkls, d_hkl in zip(xrd_pattern.x,
                                                                                         xrd_pattern.y,
                                                                                         xrd_pattern.hkls,
                                                                                         xrd_pattern.d_hkls)]

        mat["xrd"][el] = el_doc


def add_thermo(mat, thermo):
    if has(thermo, "thermo.e_above_hull"):
        set_(mat, "e_above_hull", get(thermo, "thermo.e_above_hull"))

    if has(thermo, "thermo.formation_energy_per_atom"):
        set_(mat, "formation_energy_per_atom", get(
            thermo, "thermo.formation_energy_per_atom"))

    if has(thermo, "thermo.decmposes_to"):
        set_(mat, "decmposes_to", get(thermo, "thermo.decmposes_to"))


def sandbox_props(mat):
    mat["sbxn"] = mat.get("sbxn", ["core"])
    mat["sbxd"] = []

    for sbx in mat["sbxn"]:
        sbx_d = {k: get(mat, v)
                 for k, v in SANDBOXED_PROPERTIES.items() if has(mat, k)}
        sbx_d["id"] = sbx
        mat["sbxd"].append(sbx_d)


def add_icsd(mat, icsds):

    relevant_icsd = [icsd for icsd in icsds if icsd[
        "icsd_id"] in mat.get("icsd_ids", [])]
    results = []
    for icsd in relevant_icsd:
        result = {"warnings": []}
        tags = [icsd[t]
                for t in ["chem_name", "min_name"] if t in icsd and icsd[t]]
        result["exp"] = {'pressure': icsd["pressure"], 'tags': tags}
        if icsd.get('pressure', 0) > 1:
            result["warnings"].append("High pressure experimental phase.")
        results.append(result)

    if len(results) == 0:
        results.append({"warnings": ["Structure has been removed in the 2012 version of ICSD."],
                        "exp": {}})

    if mat["icsd_ids"]:
        results[0]["exp_lattice"] = mat["snl"]["lattice"]

    # Check relaxation
    orig_crystal = Structure.from_dict(mat["snl"])
    final_structure = Structure.from_dict(mat["structure"])

    try:
        analyzer = RelaxationAnalyzer(orig_crystal, final_structure)
        latt_para_percentage_changes = analyzer.get_percentage_lattice_parameter_changes()
        for l in ["a", "b", "c"]:
            change = latt_para_percentage_changes[l] * 100
            if change < latt_para_interval[0] or change > latt_para_interval[1]:
                results[0]["warnings"].append(
                    "Large change in {} lattice parameter during relaxation.".format(l))
        change = analyzer.get_percentage_volume_change() * 100
        if change < vol_interval[0] or change > vol_interval[1]:
            results[0]["warnings"].append(
                "Large change in volume during relaxation.")
    except Exception as ex:
        # print icsd_crystal.formula
        # print final_structure.formula
        print("Relaxation analyzer failed for Material:{} due to {}".format(
            mat["task_id"], traceback.print_exc()))

    # Merge all the results

    if len(results) > 1:

        results_union = {"exp": {"tags": []}, "warnings": []}
        tags = []
        warnings = []
        for i in results:
            if i.get("exp_lattice"):
                results_union["exp_lattice"] = i["exp_lattice"]
            if i.get("exp"):
                if i["exp"].get("tags"):
                    tags.extend(i["exp"]["tags"])
            if i.get("warnings"):
                warnings.extend(i["warnings"])
        if tags:
            results_union["exp"]["tags"] = [i for i in set(tags)]
        if warnings:
            high_pressure_count = 0
            for w in warnings:
                if w == "High pressure experimental phase.":
                    high_pressure_count += 1
            results_union["warnings"] = [i for i in set(warnings)]
            if high_pressure_count != 0 and high_pressure_count != len(mat.get("icsd_ids", [])):
                results_union["warnings"].remove(
                    "High pressure experimental phase.")
        results = results_union
    else:
        results = results[0]

    if results.get("exp_lattice"):
        mat["exp_lattice"] = results["exp_lattice"]
    mat["exp"] = results["exp"]
    mat["warnings"] = results["warnings"]


def add_magnetism(mat, mag=None):
    mag_types = {"NM": "Non-magnetic",
                 "FiM": "Ferri",
                 "AFM": "AFM",
                 "FM": "FM"}

    struc = Structure.from_dict(mat["structure"])
    msa = CollinearMagneticStructureAnalyzer(struc)
    mat["magnetic_type"] = mag_types[msa.ordering.value]


def add_elasticity(mat, elasticity=None):
    elasticity = elasticity if elasticity else None
    if "elasticity" in elasticity:
        mat["elasticity"] = elasticity.get("elasticity")


def add_piezo(mat, piezo=None):
    piezo = piezo if piezo else None
    if "piezo" in piezo:
        mat["piezo"] = piezo.get("piezo")


def add_diel(mat, diel=None):
    diel = diel if diel else None
    if "diel" in diel:
        mat["diel"] = diel.get("diel")


def add_snl(mat, snl=None):
    snl = snl if snl else {}
    mat["snl"] = snl.get("snl", None)
    mat["snl_final"] = snl.get("snl", None)
    mat["created_at"] = get(snl, "snl.about.created_at") if has(
        snl, "snl.about.created_at") else datetime.utcnow()
    mat["icsd_ids"] = snl.get("icsd_ids", [])



from datetime import datetime

from maggma.builder import Builder
from pydash import py_
from tqdm import tqdm


class CopyBuilder(Builder):
    """Sync a source with a target.

    Uses `lu_field` of source and target Store to get new/updated documents.

    Can override `target.key` field to filter for target document to update.

    """

    def __init__(self, source, target, key=None, **kwargs):
        self.source = source
        self.target = target
        self.key = key if key else target.key
        super().__init__(sources=[source], targets=[target], **kwargs)

    def get_items(self):
        source, target = self.source, self.target
        lu_filter = source.lu_filter(target)
        self.logger.debug("lu_filter: {}".format(lu_filter))
        self.confirm_lu_field_index()
        cursor = source.query(criteria=lu_filter,
                              sort=[(source.lu_field, 1)])
        self.logger.info("Will copy {} items".format(cursor.count()))
        return tqdm(cursor, total=cursor.count())

    def process_item(self, item):
        return item

    def update_targets(self, items):
        source, target = self.source, self.target
        for item in items:
            # Use source last-updated value, ensuring `datetime` type.
            item[target.lu_field] = source.lu_key[0](item[source.lu_field])
            if source.lu_field != target.lu_field:
                del item[source.lu_field]
        target.update(items, update_lu=False, key=self.key)

    def confirm_lu_field_index(self):
        """Confirm index on `lu_field`.

        One can't simply ensure an index exists via
        `self.source.collection.create_index` because a Builder must assume
        read-only access to source Stores. The MongoDB `read` built-in role
        does not include the `createIndex` action.

        Raises:
            Exception: If there is no index on `lu_field`.

        """
        source = self.source
        info = source.collection.index_information().values()
        for spec in (index['key'] for index in info):
            if spec[0][0] == source.lu_field:
                break
        else:
            raise Exception("Need index on '{}' for {}".format(
                source.lu_field, source.collection))



import logging
from datetime import datetime

from monty.json import jsanitize
from monty.tempfile import ScratchDir
from pymatgen.core.structure import Structure
from pymatgen.electronic_structure.boltztrap import BoltztrapRunner

from maggma.builder import Builder

__author__ = "Shyam Dwaraknath <shyamd@lbl.gov>"


class BoltztrapBuilder(Builder):

    def __init__(self, materials, boltztrap, bandstructure_fs="bandstructure_fs", bta_fs=None, query={}, **kwargs):
        """
        Calculates conducitivty parameters using BoltzTrap
        Saves the boltztrap analyzer in bta_fs if set otherwise doesn't store it
        because it is too large usually to store in Mongo

        Args:
            materials (Store): Store of materials documents
            boltztrap (Store): Store of boltztrap
            bandstructure_fs (str): Name of the GridFS where bandstructures are stored
            query (dict): dictionary to limit materials to be analyzed

        """

        self.materials = materials
        self.boltztrap = boltztrap
        self.bandstructure_fs = bandstructure_fs
        self.bta_fs = bta_fs
        self.query = query

        super().__init__(sources=[materials],
                         targets=[boltztrap],
                         **kwargs)

    def get_items(self):
        """
        Gets all materials that need a new XRD

        Returns:
            generator of materials to calculate xrd
        """

        self.logger.info("BoltzTrap Builder Started")

        # All relevant materials that have been updated since boltztrap was last run
        # and a uniform bandstructure exists
        q = dict(self.query)
        q.update(self.materials.lu_filter(self.boltztrap))
        q["bandstructure.uniform_oid"] = {"$exists": 1}
        q["output.bandgap"] = {"$gt": 0.0}
        mats = set(self.materials.distinct(self.materials.key, criteria=q))

        # initialize the gridfs
        bfs = gridfs.GridFS(self.materials.database, self.bandstructure_fs)

        self.logger.info(
            "Found {} new materials for calculating boltztrap conductivity".format(len(mats)))
        for m in mats:
            mat = self.materials.query(
                [self.materials.key, "structure", "input.parameters.NELECT", "bandstructure"], criteria={self.materials.key: m})

            # If a bandstructure oid exists
            if "uniform_bs_oid" in mat.get("bandstructure", {}):
                bs_json = bfs.get(mat["bandstructure"][
                                  "uniform_bs_oid"]).read()

                if "zlib" in mat["bandstructure"].get("uniform_bs_compression", ""):
                    bs_json = zlib.decompress(bs_json)

                bs_dict = json.loads(bs_json.decode())
                mat["bandstructure"]["uniform_bs"] = bs_dict

            yield mat

    def process_item(self, item):
        """
        Calculates diffraction patterns for the structures

        Args:
            item (dict): a dict with a material_id and a structure

        Returns:
            dict: a diffraction dict
        """
        self.logger.debug(
            "Calculating Boltztrap for {}".format(item[self.materials.key]))

        nelect = item["input"]["parameters"]["NELECT"]

        bs_dict = item["uniform_bandstructure"]["bs"]
        bs_dict['structure'] = item['structure']
        bs = BandStructure.from_dict(bs_dict)

        with ScratchDir("."):
            BoltztrapRunner(bs=bs, nelec=nelect).run(path_dir=os.getcwd())
            btrap_dir = os.path.join(os.getcwd(), "boltztrap")
            bta = BoltztrapAnalyzer.from_files(btrap_dir)

        d = {"bta": bta.as_dict(),
             "boltztrap": {"thermoelectric": bt_analysis_thermoelectric(bta),
                           "tcm": bt_analysis_tcm(bta)}}

        return d

    def update_targets(self, items):
        """
        Inserts the new task_types into the task_types collection

        Args:
            items ([[dict]]): a list of list of thermo dictionaries to update
        """
        items = list(filter(None, items))

        bta_fs = gridfs.GridFS(self.materials.database,
                               self.bta_fs) if self.bta_fs else None

        if len(items) > 0:
            self.logger.info("Updating {} boltztrap docs".format(len(items)))

            for doc in items:
                if self.bta_fs:
                    bta_doc = dict(doc["bta"])
                    bta_json = json.dumps(jsanitize(bta_doc))
                    bta_gz = zlib.compress(bta_json)
                    bs_oid = bta_fs.put(bta_gz)
                    doc['bta_oid'] = bta_oid
                    doc['bta_compression'] = "zlib"

                del doc["bta"]

            self.boltztrap.update(items)

        else:
            self.logger.info("No items to update")


def bt_analysis_thermoelectric(bta):
    """
    Performs analysis for thermoelectrics search
    :param bta: Boltztrap analyzer object
    :return: dict of Zt,Power Factor, Seebeck, Conducitity and Kappa
    """
    d = {}

    d["zt"] = bta.get_extreme("zt")
    d["pf"] = bta.get_extreme("power factor")
    d["seebeck"] = bta.get_extreme("seebeck")
    d["conductivity"] = bta.get_extreme("conductivity")
    d["kappa_max"] = bta.get_extreme("kappa")
    d["kappa_min"] = bta.get_extreme("kappa", maximize=False)

    return d


def bt_analysis_tcm(bta, temp_min=300, temp_max=400, doping_min=1e19, doping_max=1e22):
    """
    Performs analysis for transparent conductive materials
    Focuses on T=300-400K and Doping=1E19-1E22
    :param bta: Boltztrap analyzer object
    :return: dict of conductivity and effective mass
    """
    d = {}

    d['avg_eff_mass'] = bta.get_average_eff_mass()
    d['doping'] = bta.doping

    return d



import numpy as np

from pymatgen.core.structure import Structure
import pymatgen.analysis
from pymatgen.analysis.local_env import *
from matminer.featurizers.site import OPSiteFingerprint ,\
CrystalSiteFingerprint, CoordinationNumber

# TODO:
# 1) Add checking OPs present in current implementation of site fingerprints.
# 2) Complete documentation!!!

from maggma.builder import Builder

__author__ = "Nils E. R. Zimmermann <nerz@lbl.gov>"


cls_to_abbrev = {
    'OPSiteFingerprint': 'opsf', 'CrystalSiteFingerprint': 'csf', \
    'VoronoiNN': 'vnn', 'JMolNN': 'jmnn', 'MinimumDistanceNN': 'mdnn', \
    'MinimumOKeeffeNN': 'moknn', 'MinimumVIRENN': 'mvirenn', \
    'BrunnerNN': 'bnn'}

class SiteDescriptorsBuilder(Builder):

    def __init__(self, materials, site_descriptors, mat_query=None, **kwargs):
        """
        Calculates site-based descriptors (e.g., coordination numbers
        with different near-neighbor finding approaches) for materials and
        runs statistics analysis on selected descriptor types
        (order parameter-based site fingerprints).  The latter is
        useful as a definition of a structure fingerprint
        on the basis of local coordination information.

        Args:
            materials (Store): Store of materials documents.
            site_descriptors (Store): Store of site-descriptors data such
                                      as tetrahedral order parameter or
                                      fraction of being 8-fold coordinated.
            mat_query (dict): dictionary to limit materials to be analyzed.
        """

        self.materials = materials
        self.site_descriptors = site_descriptors
        self.mat_query = mat_query if mat_query else {}

        # Set up all targeted site descriptors.
        self.sds = {}
        for nn in NearNeighbors.__subclasses__():
            nn_ = getattr(pymatgen.analysis.local_env, nn.__name__)
            t = nn.__name__ if nn.__name__ \
                not in cls_to_abbrev.keys() \
                else cls_to_abbrev[nn.__name__]
            k = 'cn_{}'.format(t)
            self.sds[k] = CoordinationNumber(nn_(), use_weights=False)
            k = 'cn_wt_{}'.format(t)
            self.sds[k] = CoordinationNumber(nn_(), use_weights=True)
        self.all_output_pieces = {'site_descriptors': [k for k in self.sds.keys()]}
        self.sds['opsf'] = OPSiteFingerprint()
        self.sds['csf'] = CrystalSiteFingerprint.from_preset('ops')
        self.all_output_pieces['statistics'] = ['opsf', 'csf']

        super().__init__(sources=[materials],
                         targets=[site_descriptors],
                         **kwargs)

    def get_items(self):
        """
        Gets all materials that need new site descriptors.
        For example, entirely new materials and materials
        for which certain descriptor in the current Store
        are still missing.

        Returns:
            generator of materials to calculate site descriptors
            and of the target quantities to be calculated
            (e.g., CN with the minimum distance near neighbor
            (MinimumDistanceNN) finding class from pymatgen which has label
            "cn_mdnn").
        """

        self.logger.info("Site-Descriptors Builder Started")

        self.logger.info("Setting indexes")

        # All relevant materials that have been updated since site-descriptors
        # were last calculated

        q = dict(self.mat_query)
        all_task_ids = list(self.materials.distinct(self.materials.key, q))
        q.update(self.materials.lu_filter(self.site_descriptors))
        new_task_ids = list(self.materials.distinct(self.materials.key, q))
        self.logger.info(
            "Found {} entirely new materials for site-descriptors data".format(
            len(new_task_ids)))
        for task_id in all_task_ids:
            if task_id in new_task_ids:
                any_piece = True

            else: # Any piece of info missing?
                data_present = self.site_descriptors.query(
                        properties=[self.site_descriptors.key, "site_descriptors", "statistics"],
                        criteria={self.site_descriptors.key: task_id}).limit(1)[0]
                any_piece = False
                for k, v in self.all_output_pieces.items():
                    if k not in list(data_present.keys()):
                        any_piece = True
                        break
                    else:
                        any_piece = False
                        for e in v:
                            if e not in data_present[k]:
                                any_piece = True
                                break
                if not any_piece:
                    for fp in ['opsf', 'csf']:
                        for l in self.sds[fp].feature_labels():
                            for fpi in data_present['site_descriptors'][fp]:
                                if l not in fpi.keys():
                                    any_piece = True
                                    break
            if any_piece:
                yield self.materials.query(
                        properties=[self.materials.key, "structure"],
                        criteria={self.materials.key: task_id}).limit(1)[0]

    def process_item(self, item):
        """
        Calculates all site descriptors for the structures


        Args:
            item (dict): a dict with a task_id and a structure

        Returns:
            dict: a site-descriptors dict
        """
        self.logger.debug("Calculating site descriptors for {}".format(
            item[self.materials.key]))

        struct = Structure.from_dict(item['structure'])

        site_descr_doc = {'structure': struct.copy()}
        site_descr_doc['site_descriptors'] = \
                self.get_site_descriptors_from_struct(
                site_descr_doc['structure'])
        site_descr_doc['statistics'] = \
                self.get_statistics(
                site_descr_doc['site_descriptors'])
        site_descr_doc[self.site_descriptors.key] = item[self.materials.key]

        return site_descr_doc

    def update_targets(self, items):
        """
        Inserts the new task_types into the task_types collection.

        Args:
            items ([[dict]]): a list of list of site-descriptors dictionaries to update.
        """
        items = list(filter(None, items))

        if len(items) > 0:
            self.logger.info("Updating {} site-descriptors docs".format(len(items)))
            self.site_descriptors.update(docs=items)
        else:
            self.logger.info("No items to update")

    def get_site_descriptors_from_struct(self, structure):
        doc = {}

        # Compute descriptors.
        for k, sd in self.sds.items():
            try:
                d = []
                l = sd.feature_labels()
                for i, s in enumerate(structure.sites):
                    d.append({'site': i})
                    for j, desc in enumerate(sd.featurize(structure, i)):
                        d[i][l[j]] = desc
                doc[k] = d

            except Exception as e:
                self.logger.error("Failed calculating {} site-descriptors: "
                                  "{}".format(k, e))

        return doc

    def get_statistics(self, site_descr, fps=('opsf', 'csf')):
        doc = {}

        # Compute site-descriptor statistics.
        for fp in fps:
            doc[fp] = {}
            try:
                n_site = len(site_descr[fp])
                tmp = {}
                for isite in range(n_site):
                    for l, v in site_descr[fp][isite].items():
                        if l not in list(tmp.keys()):
                            tmp[l] = []
                        tmp[l].append(v)
                d = []
                for k, l in tmp.items():
                    dtmp = {'name': k}
                    dtmp['min'] = min(tmp[k])
                    dtmp['max'] = max(tmp[k])
                    dtmp['mean'] = np.mean(tmp[k])
                    dtmp['std'] = np.std(tmp[k])
                    d.append(dtmp)
                doc[fp] = d

            except Exception as e:
                self.logger.error("Failed calculating statistics of site "
                                  "descriptors: {}".format(e))

        return doc



from itertools import chain
from collections import defaultdict

from pymatgen import Structure
from pymatgen.analysis.structure_matcher import StructureMatcher, ElementComparator
from pymatgen.util.provenance import StructureNL
from maggma.builder import Builder
from pydash.objects import get


class SNLBuilder(Builder):
    """
    Builds SNL collection for materials
    """

    def __init__(self, materials, snls, *source_snls, query=None, ltol=0.2, stol=0.3,
                 angle_tol=5, **kwargs):
        self.materials = materials
        self.snls = snls
        self.source_snls = list(source_snls)
        self.ltol = ltol
        self.stol = stol
        self.angle_tol = angle_tol
        self.query = query if query else {}
        self.kwargs = kwargs

        super(SNLBuilder, self).__init__(sources=[materials, *self.source_snls], targets=[snls], **kwargs)

    def get_items(self):
        """
        Gets all materials to assocaite with SNLs

        Returns:
            generator of materials and SNLs that could match
        """
        self.logger.info("SNL Builder Started")

        self.logger.info("Setting indexes")

        # Find all formulas for materials that have been updated since this
        # builder was last ran
        q = dict(self.query)
        q.update(self.materials.lu_filter(self.snls))
        forms_to_update = set(self.materials.distinct("pretty_formula", q))
        #forms_to_update = set()

        # Find all new SNL formulas since the builder was last run
        # for source in self.source_snls:
        #    new_q = source.lu_filter(self.snls)
        #    forms_to_update |= set(source.distinct("reduced_cell_formula", new_q))

        self.logger.info(
            "Found {} new/updated systems to proces".format(len(forms_to_update)))

        for formula in forms_to_update:
            mats = list(self.materials.query(properties=[
                        self.materials.key, "structure", "initial_structure", "pretty_formula"], criteria={"pretty_formula": formula}))
            snls = []

            for source in self.source_snls:
                snls.extend(source.query(criteria={"reduced_cell_formula": formula}))

            #snls = [s["snl"] for s in snls]
            self.logger.debug("Found {} snls and {} mats".format(len(snls), len(mats)))
            if len(mats) > 0 and len(snls) > 0:
                yield mats, snls

    def process_item(self, item):
        """
        Calculates diffraction patterns for the structures

        Args:
            item (tuple): a tuple of materials and snls

        Returns:
            list(dict): a list of collected snls with material ids
            """
        mats = item[0]
        source_snls = item[1]
        snls = defaultdict(list)
        self.logger.debug("Tagging SNLs for {}".format(
            mats[0]["pretty_formula"]))

        for snl in source_snls:
            mat_id = self.match(snl, mats)
            if mat_id is not None:
                snls[mat_id].append(snl)

        return snls

    def match(self, snl, mats):
        """
        Finds a material doc that matches with the given snl

        Args:
            snl (dict): the snl doc
            mats ([dict]): the materials docs to match against

        Returns:
            dict: a materials doc if one is found otherwise returns None
        """
        sm = StructureMatcher(ltol=self.ltol, stol=self.stol, angle_tol=self.angle_tol,
                              primitive_cell=True, scale=True,
                              attempt_supercell=False, allow_subset=False,
                              comparator=ElementComparator())
        snl_struc = StructureNL.from_dict(snl).structure

        for m in mats:
            m_struct = Structure.from_dict(m["structure"])
            init_m_struct = Structure.from_dict(m["initial_structure"])
            if sm.fit(m_struct, snl_struc) or sm.fit(init_m_struct, snl_struc):
                return m[self.materials.key]

        return None

    def update_targets(self, items):
        """
        Inserts the new task_types into the task_types collection

        """

        snls = [snl for snl_dict in items for snl in self.collect_snls_mp(snl_dict)]

        if len(snls) > 0:
            self.logger.info("Found {} SNLs to update".format(len(snls)))
            self.snls.update(snls)
        else:
            self.logger.info("No items to update")

    def collect_snls_mp(self, snl_dict):
        """
        Converts a dict of materials and snls into docs for the snl by choosing the first by creation date and storing all applicable ICSD ids
        """

        snls = []

        for mat_id, snl_list in snl_dict.items():
            snl = sorted(
                snl_list, key=lambda x: StructureNL.from_dict(x).created_at)[0]
            icsd_ids = list(filter(None, [get(snl, "about._icsd.icsd_id", None) for snl in snl_list]))
            snls.append({self.snls.key: mat_id, "snl": snl, "icsd_ids": icsd_ids})
        return snls



import os

from pymatgen.analysis.elasticity.elastic import ElasticTensor
from pymatgen.analysis.substrate_analyzer import SubstrateAnalyzer
from pymatgen.core import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from maggma.builder import Builder

from emmet.common.utils import load_settings
__author__ = "Shyam Dwaraknath <shyamd@lbl.gov>"


module_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))
default_substrate_settings = os.path.join(
    module_dir, "settings", "substrates.yaml")


class SubstrateBuilder(Builder):

    def __init__(self, materials, substrates, elasticity=None, substrate_settings=None,
                 query=None, **kwargs):
        """
        Calculates matching substrates for 

        Args:
            materials (Store): Store of materials documents
            diffraction (Store): Store of substrate matches
            elasticity (Store): Store of elastic tensor documents
            substrate_settings (path): Store of xrd settings
            query (dict): dictionary to limit materials to be analyzed
        """

        self.materials = materials
        self.substrates = substrates
        self.elasticity = elasticity
        self.substrate_settings = substrate_settings
        self.query = query if query else {}
        self.__settings = load_settings(
            self.substrate_settings, default_substrate_settings)

        super().__init__(sources=[materials],
                         targets=[substrates],
                         **kwargs)

    def get_items(self):
        """
        Gets all materials that need new substrates

        Returns:
            generator of materials to calculate substrates
        """

        self.logger.info("Substrate Builder Started")

        self.logger.info("Getting substrate structures")
        self.load_substrate_mats()

        e_tensor_updated_mats = self.get_mats_w_updated_elastic_tensors()
        updated_mats = self.get_updated_mats()
        updated_substrates = self.get_updated_substrates()

        mats = set(e_tensor_updated_mats + updated_mats)
        self.logger.info(
            "Updating all substrate calculations for {} materials".format(len(mats)))
        self.yield_mats(mats, self.substrate_mats)

        self.logger.info(
            "Updating remaining materials for {} new/updated substrates".format(len(updated_substrates)))
        updated_substrate_mats = [s for s in self.substrate_mats if s[
            self.materials.key] in updated_substrates]
        remaining_mats = set(self.materials.distinct(
            self.materials.key, self.query)) - set(mats)
        self.yield_mats(remaining_mats, updated_substrate_mats)

    def process_item(self, item):
        """
        Calculates substrate matches for all given substrates

        Args:
            item (dict): a dict with a material_id and a structure

        Returns:
            dict: a diffraction dict
        """
        material = item["material"]
        elastic_tensor = item.get("elastic_tensor", None)
        substrates = item["substrates"]

        self.logger.debug("Calculating substrates for {}".format(
            get(self.materials.key, material)))

        film = Structure.from_dict(material["structure"])

        all_matches = []

        sa = SubstrateAnalyzer()

        for s in substrates:

            substrate = Structure.from_dict(s["structure"])

            # Calculate all matches and group by substrate orientation
            matches_by_orient = groupby_itemkey(
                sa.calculate(film, substrate, elastic_tensor, lowest=True),
                "sub_miller")

            # Find the lowest area match for each substrate orientation
            lowest_matches = [min(g, key=itemgetter("match_area"))
                              for k, g in matches_by_orient]

            for match in lowest_matches:
                db_entry = {
                    "film_id": material[self.materials.key],
                    "sub_id": s[self.materials.key],
                    "orient": " ".join(map(str, match["sub_miller"])),
                    "sub_form": substrate.composition.reduced_formula,
                    "film_orient": " ".join(map(str, match["film_miller"])),
                    "area": match["match_area"],
                }
                if "elastic_energy" in match:
                    db_entry["energy"] = match["elastic_energy"]
                    db_entry["strain"] = match["strain"]

                all_matches.append(db_entry)

    return all_matches

    def update_targets(self, items):
        """
        Inserts the new substrate matches into the substrates collection

        Args:
            items ([[dict]]): a list of list of thermo dictionaries to update
        """

        items = list(filter(None, chain.from_iterable(items)))

        if len(items) > 0:
            self.logger.info(
                "Updating {} substrate matches".format(len(items)))
            self.substrates.update(
                key=["film_id", "sub_id", "orient"], docs=items)
        else:
            self.logger.info("No items to update")

    def load_substrate_mats(self):
        """
        Loads the substrate structures from the materials collection
        """

        # TODO: Switch this to be able to take a list of structures and match materials
        # Much more usefull if you want to just prescribe ICSD structures for
        # instance
        mats = [self.materials.find_one(properties=["structure", self.materials.key, self.materials.lu_key], criteria={
                                        self.materials.key: mpid}) for mpid in self.__settings]
        self.substrate_mats = list(filter(None, mats))

    def get_mats_w_updated_elastic_tensors(self):
        e_tensor_updated_mats = []
        self.logger.info("Checking for new/updated elastic tensors")
        if self.elasticity:
            # Find all materials with updated elastic tensors since the
            # substrate builder was last run and rerun all substrates for
            # analysis
            q = self.elasticity.lu_filter(self.substrates)
            e_tensor_updated_mats = self.elasticity.distinct(
                self.elasticity.key, criteria=q)
            # Ensure these materials are within our materials query
            q = dict(query)
            q.update({self.materials.key: {"$in": e_tensor_updated_mats}})
            e_tensor_updated_mats = self.materials.distinct(
                self.materials.key, criteria=q)
            self.logger.info(
                "Found {} new/updated elastic tensors".format(len(e_tensor_updated_mats)))

        return e_tensor_updated_mats

    def get_updated_mats(self):
        self.logger.info("Checking for new/updated materials")
        q = dict(self.query)
        q.update(self.materials.lu_filter(self.substrates))
        updated_mats = list(self.materials.distinct(self.materials.key, q))
        self.logger.info(
            "Found {} new materials for substrate processing".format(len(updated_mats)))
        return updated_mats

    def get_updated_substrates(self):
        # TODO: How do we want to do this?
        self.logger.info("Checking for new/updated substrates")
        for s in self.substrate_mats:
            pass
        return []

    def yield_mats(self, mats, substrates):
        for m in mats:
            e_tensor = self.elasticity.find_one(
                criteria={self.elasticity.key: m})
            mat = self.materials.find_one(
                props=["structure", self.materials.key], criteria={self.materials.key: m})
            yield {"material": mat,
                   "elastic_tensor": e_tensor,
                   "substrates": substrates}


def conventional_standard_structure(doc):
    """Get a conventional standard structure from doc["structure"]."""
    s = Structure.from_dict(doc["structure"])
    spga = SpacegroupAnalyzer(s, symprec=0.1)
    return spga.get_conventional_standard_structure()


def groupby_itemkey(iterable, item):
    """groupby keyed on (and pre-sorted by) itemgetter(item)."""
    itemkey = itemgetter(item)
    return itertools.groupby(sorted(iterable, key=itemkey), itemkey)



from monty.serialization import loadfn

def load_settings(settings,default_settings):
	if os.path.is_path(settings):
		return loadfn(settings)
	elif type(settings) is dict or type(settings) is list:
		return settings
	else:
		return loadfn(default_settings)


import numpy as np
from math import sqrt

from maggma.builder import Builder

__author__ = "Nils E. R. Zimmermann <nerz@lbl.gov>"


class StructureSimilarityBuilder(Builder):

    def __init__(self, site_descriptors, structure_similarity,
                 fp_type='csf', **kwargs):
        """
        Calculates similarity metrics between structures on the basis
        of site descriptors.

        Args:
            site_descriptors (Store): storage of site-descriptors data
                                      such as tetrahedral order parameter
                                      or percentage of 8-fold coordination.
            structure_similarity (Store): storage of structure similarity
                                          metrics.
            fp_type (str): target site fingerprint type to be
                           used for similarity computation
                           ("csf" (based on matminer's
                           CrystalSiteFingerprint class)
                           or "opsf" (based on matminer's
                           OPSiteFingerprint class)).
        """

        self.site_descriptors = site_descriptors
        self.structure_similarity = structure_similarity
        self.fp_type = fp_type

        super().__init__(sources=[site_descriptors],
                         targets=[structure_similarity],
                         **kwargs)

    def get_items(self):
        """
        Gets all materials that need new site descriptors.

        Returns:
            generator of materials to calculate site descriptors.
        """

        self.logger.info("Structure Similarity Builder Started")

        self.logger.info("Setting indexes")

        #TODO: re-introduce last-updated filtering.
        task_ids = list(self.site_descriptors.distinct(self.site_descriptors.key))
        n_task_ids = len(task_ids)
        for i in range(n_task_ids-1):
            d1 = self.site_descriptors.query(
                    properties=[self.site_descriptors.key, "statistics"],
                    criteria={self.site_descriptors.key: task_ids[i]}).limit(1)[0]
            for j in range(i+1, n_task_ids):
                d2 = self.site_descriptors.query(
                        properties=[
                        self.site_descriptors.key, "statistics"],
                        criteria={self.site_descriptors.key: task_ids[j]}).limit(1)[0]
                yield list([d1, d2])

    def process_item(self, item):
        """
        Calculates site descriptors for the structures

        Args:
            item (list): a list (length 2) with each one document that
                         carries a task ID in "task_id" and a statistics
                         vector from OP site-fingerprints in
                         "statistics".

        Returns:
            dict: similarity measures.
        """
        self.logger.debug("Similarities for {} and {}".format(
            item[0][self.site_descriptors.key],
            item[1][self.site_descriptors.key]))

        sim_doc = {}
        sim_doc = self.get_similarities(
            item[0], item[1])
        sim_doc[self.structure_similarity.key] = tuple(
            sorted([item[0][self.site_descriptors.key],
            item[1][self.site_descriptors.key]]))

        return sim_doc

    def update_targets(self, items):
        """
        Inserts the new task_types into the task_types collection.

        Args:
            items ([[dict]]): a list of list of site-descriptors dictionaries to update.
        """
        if len(items) > 0:
            self.logger.info("Updating {} structure-similarity docs".format(len(items)))
            self.structure_similarity.update(docs=items)
        else:
            self.logger.info("No items to update")

    def get_similarities(self, d1, d2):
        doc = {}

        # Compute similarty metrics.
        try:
            dout = {}
            l = {}
            v = {}
            for i, li in enumerate([d1['statistics'][self.fp_type],
                                  d2['statistics'][self.fp_type]]):
                v[i] = []
                l[i] = []
                #for optype, stats in d.items():
                for opdict in li:
                    for stattype, val in opdict.items():
                        if stattype != 'name':
                            v[i].append(val)
                            l[i].append('{} {}'.format(opdict['name'], stattype))
            if len(l[0]) != len(l[1]):
                raise RuntimeError('Site-fingerprint statistics dictionaries'
                                   ' have different sizes ({}, {})'.format(
                                   len(l[0]), len(l[1])))
            for k in l[0]:
                if k not in l[1]:
                    raise RuntimeError('Label "{}" not found in second site-'
                                       'fingerprint statistics '
                                       'dictionary'.format(k))
            v1 = np.array([v[0][k] for k in range(len(l[0]))])
            v2 = np.array([v[1][l[1].index(k)] for k in l[0]])
            dout['cos'] = np.dot(v1, v2) / (
                    np.linalg.norm(v1) * np.linalg.norm(v2))
            dout['dist'] = np.linalg.norm(v1 - v2)
            doc = dout

        except Exception as e:
             self.logger.error("Failed calculating structure similarity"
                              "metrics: {}".format(e))

        return doc



import gridfs
import json
import zlib
import io
import traceback
import numpy as np
from maggma.builder import Builder
from pydash.objects import get
import prettyplotlib as ppl
import matplotlib
import scipy.interpolate as scint
from prettyplotlib import brewer2mpl
from pymatgen.core.structure import Structure
from pymatgen.symmetry.bandstructure import HighSymmKpath
from pymatgen.electronic_structure.bandstructure import BandStructureSymmLine, BandStructure
from pymatgen.electronic_structure.dos import CompleteDos
from pymatgen.electronic_structure.plotter import DosPlotter, BSPlotter
from pymatgen.util.plotting import pretty_plot

__author__ = "Shyam Dwaraknath <shyamd@lbl.gov>"

matplotlib.use('agg')


class ElectronicStructureImageBuilder(Builder):

    def __init__(self,
                 materials,
                 electronic_structure,
                 query=None,
                 **kwargs):
        """
        Creates an electronic structure from a tasks collection, the associated band structures and density of states, and the materials structure

        Really only usefull for MP Website infrastructure right now.

        materials (Store) : Store of materials documents
        electronic_structure  (Store) : Store of electronic structure documents
        query (dict): dictionary to limit tasks to be analyzed
        """

        self.materials = materials
        self.electronic_structure = electronic_structure
        self.query = query if query else {}

        super().__init__(sources=[materials], targets=[electronic_structure], **kwargs)

    def get_items(self):
        """
        Gets all items to process into materials documents

        Returns:
            generator or list relevant tasks and materials to process into materials documents
        """

        self.logger.info("Electronic Structure Builder Started")

        # only consider materials that were updated since the electronic structure was last updated
        # and there is either a dos or bandstructure
        q = dict(self.query)
        q.update(self.materials.lu_filter(self.electronic_structure))
        q["$or"] = [{"bandstructure.bs_oid": {"$exists": 1}}, {"bandstructure.dos_oid": {"$exists": 1}}]

        # initialize the gridfs
        self.bfs = gridfs.GridFS(self.materials.collection.database, "bandstructure_fs")
        self.dfs = gridfs.GridFS(self.materials.collection.database, "dos_fs")

        mats = list(self.materials.distinct(self.materials.key, criteria=q))

        self.logger.debug("Processing {} materials for electronic structure".format(len(mats)))
        for m in mats:
            mat = self.materials.query_one(
                [self.materials.key, "structure", "bandstructure", "origins", "calc_settings", "inputs"], {
                    self.materials.key: m
                })
            self.get_bandstructure(mat)
            self.get_dos(mat)

            yield mat

    def process_item(self, mat):
        """
        Process the tasks and materials into just a list of materials

        Args:
            mat (dict): material document

        Returns:
            (dict): electronic_structure document
        """
        d = {self.electronic_structure.key: mat[self.materials.key]}
        self.logger.info("Processing: {}".format(mat[self.materials.key]))

        bs = extract_bs(mat)
        dos = extract_dos(mat)
        ylim = None

        # Plot Band structure
        if bs:
            try:
                plotter = WebBSPlotter(bs)
                plot = plotter.get_plot()
                ylim = plot.ylim()
                d["bs_plot"] = image_from_plot(plot)
                plot.close()
            except Exception:
                self.logger.warning("Caught error in bandstructure plotting for {}: {}".format(
                    mat[self.materials.key], traceback.format_exc()))

        # Reduced Band structure plot
        try:
            gap = bs.get_band_gap()["energy"]
            plot_data = plotter.bs_plot_data()
            d["bs_plot_small"] = get_small_plot(plot_data, gap)
        except Exception:
            self.logger.warning("Caught error in generating reduced bandstructure plot for {}: {}".format(
                mat[self.materials.key], traceback.format_exc()))

        # Plot DOS
        if dos:
            try:
                plotter = WebDosVertPlotter()
                plotter.add_dos_dict(dos.get_element_dos())
                plot = plotter.get_plot(ylim=ylim)
                d["dos_plot"] = image_from_plot(plot)
                plot.close()
            except Exception:
                self.logger.warning("Caught error in dos plotting for {}: {}".format(
                    mat[self.materials.key], traceback.format_exc()))

        # Get basic bandgap properties
        if bs:
            try:
                d["band_gap"] = {
                    "band_gap": bs.get_band_gap()["energy"],
                    "direct_gap": bs.get_direct_band_gap(),
                    "is_direct": bs.get_band_gap()["direct"],
                    "transition": bs.get_band_gap()["transition"]
                }
            except Exception:
                self.logger.warning("Caught error in calculating bandgap {}: {}".format(mat[self.materials.key],
                                                                                        traceback.format_exc()))

        # Store OID for GridFS access from website
        if "bs_oid" in mat.get("bandstructure", {}):
            d["bs_oid"] = mat["bandstructure"]["bs_oid"]
            d["bs_compression"] = mat["bandstructure"].get("bs_compression", "")

        if "dos_oid" in mat.get("bandstructure", {}):
            d["dos_oid"] = mat["bandstructure"]["dos_oid"]
            d["dos_compression"] = mat["bandstructure"].get("dos_compression", "")
        return d

    def update_targets(self, items):
        """
        Inserts the new task_types into the task_types collection

        Args:
            items ([([dict],[int])]): A list of tuples of materials to update and the corresponding processed task_ids
        """
        items = list(filter(None, items))

        if len(items) > 0:
            self.logger.info("Updating {} electronic structure docs".format(len(items)))
            self.electronic_structure.update(items)
        else:
            self.logger.info("No electronic structure docs to update")

    #
    # These are all helper methods designed to take a chunk of code and provide a description
    #

    def get_bandstructure(self, mat):

        # If a bandstructure oid exists
        if "bs_oid" in mat.get("bandstructure", {}):
            bs_json = self.bfs.get(mat["bandstructure"]["bs_oid"]).read()

            if "zlib" in mat["bandstructure"].get("bs_compression", ""):
                bs_json = zlib.decompress(bs_json)

            bs_dict = json.loads(bs_json.decode())
            mat["bandstructure"]["bs"] = bs_dict

    def get_dos(self, mat):

        # if a dos oid exists
        if "dos_oid" in mat.get("bandstructure", {}):
            dos_json = self.dfs.get(mat["bandstructure"]["dos_oid"]).read()

            if "zlib" in mat["bandstructure"].get("dos_compression", ""):
                dos_json = zlib.decompress(dos_json)

            dos_dict = json.loads(dos_json.decode())
            mat["bandstructure"]["dos"] = dos_dict

def extract_bs(mat):

    bs = None

    # Process the bandstructure for information
    if "bs" in mat["bandstructure"]:
        bs_dict = mat["bandstructure"]["bs"]
        # Add in structure if not already there
        if "structure" not in bs_dict:
            bs_dict["structure"] = mat["structure"]

        # Add in High Symm K Path if not already there
        if len(bs_dict.get("labels_dict", {})) == 0:
            labels = get(mat, "inputs.nscf_line.kpoints.labels", None)
            kpts = get(mat, "inputs.nscf_line.kpoints.kpoints", None)
            if labels and kpts:
                labels_dict = dict(zip(labels, kpts))
                labels_dict.pop(None, None)
            else:
                struc = Structure.from_dict(mat["structure"])
                labels_dict = HighSymmKpath(struc)._kpath["kpoints"]

            bs_dict["labels_dict"] = labels_dict

        bs = BandStructureSymmLine.from_dict(BandStructure.from_dict(bs_dict).as_dict())

    return bs

def extract_dos(mat):

    dos = None
    if "dos" in mat["bandstructure"]:
        dos_dict = mat["bandstructure"]["dos"]
        dos = CompleteDos.from_dict(dos_dict)
    return dos


def get_small_plot(plot_data, gap):
    for branch in plot_data['energy']:
        for spin, v in branch.items():
            new_bands = []
            for band in v:
                if min(band) < gap + 3 and max(band) > -3:
                    new_bands.append(band)
            branch[spin] = new_bands
    return plot_data


def image_from_plot(plot):
    imgdata = io.BytesIO()
    plot.savefig(imgdata, format="png", dpi=100)
    plot_img = imgdata.getvalue()
    return plot_img


#
# Obtain web-friendly images by subclassing pymatgen plotters.
# Should eventually phase out to BSDOS Plotter
#


class WebBSPlotter(BSPlotter):

    def get_plot(self, zero_to_efermi=True, ylim=None, smooth=False):
        """
        get a matplotlib object for the bandstructure plot.
        Blue lines are up spin, red lines are down
        spin.
        Args:
            zero_to_efermi: Automatically subtract off the Fermi energy from
                the eigenvalues and plot (E-Ef).
            ylim: Specify the y-axis (energy) limits; by default None let
                the code choose. It is vbm-4 and cbm+4 if insulator
                efermi-10 and efermi+10 if metal
            smooth: interpolates the bands by a spline cubic
        """

        plt = pretty_plot(6, 5.5)  # Was 12, 8

        matplotlib.rc('text', usetex=True)

        width = 4
        ticksize = int(width * 2.5)
        axes = plt.gca()
        axes.set_title(axes.get_title(), size=width * 4)
        labelsize = int(width * 3)
        axes.set_xlabel(axes.get_xlabel(), size=labelsize)
        axes.set_ylabel(axes.get_ylabel(), size=labelsize)

        plt.xticks(fontsize=ticksize)
        plt.yticks(fontsize=ticksize)

        for axis in ['top', 'bottom', 'left', 'right']:
            axes.spines[axis].set_linewidth(0.5)

        # main internal config options
        e_min = -4
        e_max = 4
        if self._bs.is_metal():
            e_min = -10
            e_max = 10
        band_linewidth = 1  # Was 3 in pymatgen

        data = self.bs_plot_data(zero_to_efermi)
        if not smooth:
            for d in range(len(data['distances'])):
                for i in range(self._nb_bands):
                    plt.plot(
                        data['distances'][d],
                        [data['energy'][d][str(Spin.up)][i][j] for j in range(len(data['distances'][d]))],
                        'b-',
                        linewidth=band_linewidth)
                    if self._bs.is_spin_polarized:
                        plt.plot(
                            data['distances'][d],
                            [data['energy'][d][str(Spin.down)][i][j] for j in range(len(data['distances'][d]))],
                            'r--',
                            linewidth=band_linewidth)
        else:
            for d in range(len(data['distances'])):
                for i in range(self._nb_bands):
                    tck = scint.splrep(
                        data['distances'][d],
                        [data['energy'][d][str(Spin.up)][i][j] for j in range(len(data['distances'][d]))])
                    step = (data['distances'][d][-1] - data['distances'][d][0]) / 1000

                    plt.plot(
                        [x * step + data['distances'][d][0] for x in range(1000)],
                        [scint.splev(x * step + data['distances'][d][0], tck, der=0) for x in range(1000)],
                        'b-',
                        linewidth=band_linewidth)

                    if self._bs.is_spin_polarized:

                        tck = scint.splrep(
                            data['distances'][d],
                            [data['energy'][d][str(Spin.down)][i][j] for j in range(len(data['distances'][d]))])
                        step = (data['distances'][d][-1] - data['distances'][d][0]) / 1000

                        plt.plot(
                            [x * step + data['distances'][d][0] for x in range(1000)],
                            [scint.splev(x * step + data['distances'][d][0], tck, der=0) for x in range(1000)],
                            'r--',
                            linewidth=band_linewidth)
        self._maketicks(plt)

        # Main X and Y Labels
        plt.xlabel(r'$\mathrm{Wave\ Vector}$')
        ylabel = r'$\mathrm{E\ -\ E_f\ (eV)}$' if zero_to_efermi \
            else r'$\mathrm{Energy\ (eV)}$'
        plt.ylabel(ylabel)

        # Draw Fermi energy, only if not the zero
        if not zero_to_efermi:
            ef = self._bs.efermi
            plt.axhline(ef, linewidth=2, color='k')

        # X range (K)
        # last distance point
        x_max = data['distances'][-1][-1]
        plt.xlim(0, x_max)

        if ylim is None:
            if self._bs.is_metal():
                # Plot A Metal
                if zero_to_efermi:
                    plt.ylim(e_min, e_max)
                else:
                    plt.ylim(self._bs.efermi + e_min, self._bs._efermi + e_max)
            else:
                for cbm in data['cbm']:
                    plt.scatter(cbm[0], cbm[1], color='r', marker='o', s=100)

                for vbm in data['vbm']:
                    plt.scatter(vbm[0], vbm[1], color='g', marker='o', s=100)
                plt.ylim(data['vbm'][0][1] + e_min, data['cbm'][0][1] + e_max)
        else:
            plt.ylim(ylim)

        plt.tight_layout()

        return plt


class WebDosVertPlotter(DosPlotter):

    def get_plot(self, xlim=None, ylim=None, plt=None, handle_only=False):
        """
        Get a matplotlib plot showing the DOS.
        Args:
            xlim: Specifies the x-axis limits. Set to None for automatic
                determination.
            ylim: Specifies the y-axis limits.
            plt: Handle on existing plot.
            handle_only: Quickly return just a handle. Useful if this method
                raises an exception so that one can close() the figure.
        """

        plt = plt or pretty_plot(2, 5.5)
        if handle_only:
            return plt

        ncolors = max(3, len(self._doses))
        ncolors = min(9, ncolors)
        colors = brewer2mpl.get_map('Set1', 'qualitative', ncolors).mpl_colors

        y = None
        alldensities = []
        allenergies = []

        width = 4
        ticksize = int(width * 2.5)
        axes = plt.gca()
        axes.set_title(axes.get_title(), size=width * 4)
        labelsize = int(width * 3)
        axes.set_xlabel(axes.get_xlabel(), size=labelsize)
        axes.set_ylabel(axes.get_ylabel(), size=labelsize)
        axes.xaxis.labelpad = 6

        # Note that this complicated processing of energies is to allow for
        # stacked plots in matplotlib.
        for key, dos in self._doses.items():
            energies = dos['energies']
            densities = dos['densities']
            if not y:
                y = {Spin.up: np.zeros(energies.shape), Spin.down: np.zeros(energies.shape)}
            newdens = {}
            for spin in [Spin.up, Spin.down]:
                if spin in densities:
                    if self.stack:
                        y[spin] += densities[spin]
                        newdens[spin] = y[spin].copy()
                    else:
                        newdens[spin] = densities[spin]
            allenergies.append(energies)
            alldensities.append(newdens)

        keys = list(self._doses.keys())
        keys.reverse()
        alldensities.reverse()
        allenergies.reverse()
        allpts = []
        for i, key in enumerate(keys):
            x = []
            y = []
            for spin in [Spin.up, Spin.down]:
                if spin in alldensities[i]:
                    densities = list(int(spin) * alldensities[i][spin])
                    energies = list(allenergies[i])
                    if spin == Spin.down:
                        energies.reverse()
                        densities.reverse()
                    y.extend(energies)
                    x.extend(densities)
            allpts.extend(list(zip(x, y)))
            if self.stack:
                plt.fill(x, y, color=colors[i % ncolors], label=str(key))
            else:
                ppl.plot(x, y, color=colors[i % ncolors], label=str(key), linewidth=1)
            if not self.zero_at_efermi:
                xlim = plt.xlim()
                ppl.plot(
                    xlim, [self._doses[key]['efermi'], self._doses[key]['efermi']],
                    color=colors[i % ncolors],
                    linestyle='--',
                    linewidth=1)

        if ylim:
            plt.ylim(ylim)
        if xlim:
            plt.xlim(xlim)
        else:
            ylim = plt.ylim()
            relevantx = [p[0] for p in allpts if ylim[0] < p[1] < ylim[1]]
            plt.xlim(min(relevantx), max(relevantx))
        if self.zero_at_efermi:
            xlim = plt.xlim()
            plt.plot(xlim, [0, 0], 'k--', linewidth=1)

        plt.ylabel(r'$\mathrm{E\ -\ E_f\ (eV)}$')
        plt.xlabel(r'$\mathrm{Density\ of\ states}$')

        locs, _ = plt.xticks()
        plt.xticks([0], fontsize=ticksize)
        plt.yticks(fontsize=ticksize)
        plt.grid(which='major', axis='y')

        plt.legend(fontsize='x-small', loc='upper right', bbox_to_anchor=(1.15, 1))
        leg = plt.gca().get_legend()
        leg.get_frame().set_alpha(0.25)
        plt.tight_layout()
        return plt



import numpy as np
import logging
import warnings
from datetime import datetime
from itertools import chain

from monty.json import jsanitize

from pymatgen import Structure
from pymatgen.analysis.elasticity.elastic import ElasticTensor
from pymatgen.analysis.elasticity.strain import Deformation
from pymatgen.analysis.elasticity.stress import Stress
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from maggma.builder import Builder

from atomate.utils.utils import get_mongolike

import tqdm


__author__ = "Joseph Montoya, Shyam Dwaraknath"
__maintainer__ = "Joseph Montoya"
__email__ = "montoyjh@lbl.gov"


logger = logging.getLogger(__name__)


class ElasticBuilder(Builder):
    def __init__(self, tasks, elasticity, materials,
                 query=None, incremental=True, **kwargs):
        """
        Creates a elastic collection for materials

        Args:
            tasks (Store): Store of task documents
            elastic (Store): Store of elastic properties
            materials (Store): Store of materials properties
            query (dict): dictionary to limit tasks to be analyzed
            incremental (bool): whether or not to use a lu_filter based
                on the current datetime
        """

        self.tasks = tasks
        self.elasticity = elasticity
        self.materials = materials
        self.query = query if query is not None else {}
        self.incremental = incremental
        self.start_date = datetime.utcnow()

        super().__init__(sources=[tasks],
                         targets=[elasticity],
                         **kwargs)

    def connect(self):
        self.tasks.connect()
        self.elasticity.connect()
        self.materials.connect()

    def get_items(self):
        """
        Gets all items to process into materials documents

        Returns:
            generator of tasks aggregated by formula with relevant data
            projection to process into elasticity documents
        """

        self.logger.info("Elastic Builder Started")
        self.logger.debug("Adding indices")
        self.tasks.ensure_index("nsites")
        self.tasks.ensure_index("formula_pretty")

        # Get only successful elastic deformation tasks with parent structure
        q = dict(self.query)
        q["state"] = "successful"
        q["task_label"] = {"$in": ["elastic deformation",
                                   "structure optimization"]}

        return_props = ['output', 'input', 'completed_at',
                        'transmuter', 'task_id', 'task_label', 'formula_pretty']
        self.logger.debug("Getting criteria")
        formulas = self.tasks.distinct('formula_pretty', criteria=q)

        material_dict = generate_formula_dict(self.materials)

        # formulas that have been updated since elasticity was last updated
        # Note that this makes the builder a bit slower if run for a complete
        # build in non-incremental
        if self.incremental:
            self.logger.info("Ensuring indices on lu_field for sources/targets")
            self.tasks.ensure_index(self.tasks.lu_field)
            self.elasticity.ensure_index(self.elasticity.lu_field)
            incr_filter = q.copy()
            incr_filter.update(self.tasks.lu_filter(self.elasticity))
            new_formulas = self.tasks.distinct("formula_pretty", incr_filter)
            q.update({"formula_pretty": {"$in": new_formulas}})

        cmd_cursor = self.tasks.groupby("formula_pretty",
                                        properties=return_props,
                                        criteria=q)

        for n, doc in enumerate(cmd_cursor):
            # TODO: refactor for task sets without structure opt
            logger.debug("Processing formula {}, {} of {}".format(
                doc['_id'], n, len(formulas)))
            possible_mp_ids = material_dict.get(doc['_id']["formula_pretty"])
            if possible_mp_ids:
                yield doc['docs'], possible_mp_ids
            else:
                yield None, None

    def process_item(self, item):
        """
        Process the tasks and materials into a elasticity collection

        Args:
            item: a dictionary of documents keyed by materials id

        Returns:
            an elasticity document
        """
        all_docs = []
        tasks, material_dict = item
        if not tasks:
            return all_docs
        grouped = group_by_material_id(material_dict, tasks)
        for mp_id, task_sets in grouped.items():
            elastic_docs = []
            for opt_task, defo_tasks in task_sets:
                elastic_doc = get_elastic_analysis(opt_task, defo_tasks)
                if elastic_doc:
                    elastic_docs.append(elastic_doc)

            if not elastic_docs:
                logger.warning("No elastic doc for mp_id {}".format(mp_id))
                continue

            # For now just do the most recent one that's not failed
            sorted(elastic_docs, key=lambda x: (x['state'], x['completed_at']))
            final_doc = elastic_docs[-1]
            c_ijkl = ElasticTensor.from_voigt(final_doc['elastic_tensor'])
            structure = final_doc['optimized_structure']
            formula = structure.composition.reduced_formula
            elements = [s.symbol for s in structure.composition.elements]
            chemsys = '-'.join(elements)
            final_doc.update(c_ijkl.property_dict)
            final_doc.update(c_ijkl.get_structure_property_dict(structure))

            elastic_summary = {'material_id': mp_id,
                               'all_elastic_fits': elastic_docs,
                               'elasticity': final_doc,
                               'pretty_formula': formula,
                               'chemsys': chemsys,
                               'elements': elements,
                               'last_updated': self.elasticity.lu_field}
            all_docs.append(elastic_summary)
            # elastic_summary.update(final_doc)

        return all_docs


    def update_targets(self, items):
        """
        Inserts the new elasticity documents into the elasticity collection

        Args:
            items ([dict]): list of elasticity docs
        """
        items = filter(bool, items)
        items = chain.from_iterable(items)
        items = [jsanitize(doc, strict=True) for doc in items]

        self.logger.info("Updating {} elastic documents".format(len(items)))

        self.elasticity.update(items, key='material_id')

    def finalize(self, items):
        pass

def get_elastic_analysis(opt_task, defo_tasks):
    """
    Performs the analysis of opt_tasks and defo_tasks necessary for
    an elastic analysis

    Args:
        opt_task: task doc corresponding to optimization
        defo_tasks: task_doc corresponding to deformations

    Returns:
        elastic document with fitted elastic tensor and analysis

    """
    elastic_doc = {"warnings": []}
    opt_struct = Structure.from_dict(opt_task['output']['structure'])
    d_structs = [Structure.from_dict(d['output']['structure'])
                 for d in defo_tasks]
    defos = [calculate_deformation(opt_struct, def_structure)
             for def_structure in d_structs]

    # Warning if deformation is not equivalent to stored deformation
    stored_defos = [d['transmuter']['transformation_params'][0]\
                     ['deformation'] for d in defo_tasks]
    if not np.allclose(defos, stored_defos, atol=1e-5):
        wmsg = "Inequivalent stored and calc. deformations."
        logger.warning(wmsg)
        elastic_doc["warnings"].append(wmsg)

    # Collect all fitting data and task ids
    defos = [Deformation(d) for d in defos]
    strains = [d.green_lagrange_strain for d in defos]
    vasp_stresses = [d['output']['stress'] for d in defo_tasks]
    cauchy_stresses = [-0.1 * Stress(s) for s in vasp_stresses]
    pk_stresses = [Stress(s.piola_kirchoff_2(d))
                   for s, d in zip(cauchy_stresses, defos)]
    defo_task_ids = [d['task_id'] for d in defo_tasks]

    # Determine whether data is sufficient to fit tensor
    # If raw data is insufficient but can be symmetrically transformed
    # to provide a sufficient set, use the expanded set with appropriate
    # symmetry transformations, fstresses/strains are "fitting
    # strains" below.
    vstrains = [s.voigt for s in strains]
    if np.linalg.matrix_rank(vstrains) < 6:
        symmops = SpacegroupAnalyzer(opt_struct).get_symmetry_operations()
        fstrains = [[s.transform(symmop) for symmop in symmops]
                    for s in strains]
        fstrains = list(chain.from_iterable(fstrains))
        vfstrains = [s.voigt for s in fstrains]
        if not np.linalg.matrix_rank(vfstrains) == 6:
            logger.warning("Insufficient data to form SOEC")
            elastic_doc['warnings'].append("insufficient strains")
            return None
        else:
            fstresses = [[s.transform(symmop) for symmop in symmops]
                         for s in pk_stresses]
            fstresses = list(chain.from_iterable(fstresses))
    else:
        fstrains = strains
        fstresses = pk_stresses

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        if len(cauchy_stresses) == 24:
            elastic_doc['legacy_fit'] = legacy_fit(strains, cauchy_stresses)
        et_raw = ElasticTensor.from_pseudoinverse(fstrains, fstresses)
        et = et_raw.voigt_symmetrized.convert_to_ieee(opt_struct)
        defo_tasks = sorted(defo_tasks, key=lambda x: x['completed_at'])
        vasp_input = opt_task['input']
        vasp_input.pop('structure')

        elastic_doc.update({"deformation_task_ids": defo_task_ids,
                            "optimization_task_id": opt_task['task_id'],
                            "pk_stresses": pk_stresses,
                            "cauchy_stresses": cauchy_stresses,
                            "strains": strains,
                            "deformations": defos,
                            "elastic_tensor": et.voigt,
                            "elastic_tensor_raw": et_raw.voigt,
                            "optimized_structure": opt_struct,
                            "completed_at": defo_tasks[-1]['completed_at'],
                            "optimization_input": vasp_input})

    # Process input
    elastic_doc['warnings'] = get_warnings(et, opt_struct) or None
    #TODO: process MPWorks metadata?
    #TODO: higher order
    #TODO: add some of the relevant DFT params, kpoints
    elastic_doc['state'] = "filter_failed" if elastic_doc['warnings']\
        else "successful"
    return elastic_doc


def group_by_material_id(materials_dict, docs, tol=1e-6,
                         structure_matcher=None):
    """
    Groups a collection of documents by material id
    as found in a materials collection

    Args:
        materials_dict (dict): dictionary of structures keyed by material_id
        docs ([dict]): list of documents
        tol: tolerance for lattice grouping
        structure_matcher (StructureMatcher): structure
            matcher for finding equivalent structures

    Returns:
        documents grouped by material_id from the materials
        collection
    """
    sm = structure_matcher or StructureMatcher()
    tasks_by_opt = group_deformations_by_optimization_task(docs, tol)
    task_sets_by_mp_id = {}
    for opt_task, defo_tasks in tasks_by_opt:
        structure = Structure.from_dict(opt_task['output']['structure'])
        match = False
        for c_id, candidate in materials_dict.items():
            c_structure = Structure.from_dict(candidate)
            if sm.fit(c_structure, structure):
                mp_id = c_id
                match = True
                break
        if match:
            if mp_id in task_sets_by_mp_id:
                task_sets_by_mp_id[mp_id].append((opt_task, defo_tasks))
            else:
                task_sets_by_mp_id[mp_id] = [(opt_task, defo_tasks)]
    return task_sets_by_mp_id


def group_deformations_by_optimization_task(docs, tol=1e-6):
    """
    Groups a set of deformation tasks by equivalent lattices
    to an optimization task.  Basically the same as
    group_by_parent_lattice, except does an additional
    step of finding the optimization and using that
    as the grouping parameter.  Also filters document
    sets that don't include an optimization and deformations.

    Args:
        docs ([{}]): list of documents
        tol (float): tolerance for lattice equivalence
    """
    # TODO: this could prolly be refactored to be more generally useful
    tasks_by_lattice = group_by_parent_lattice(docs, tol)
    tasks_by_opt_task = []
    for _, task_set in tasks_by_lattice:
        opt_struct_tasks = [task for task in task_set
                            if task['task_label']=='structure optimization']
        deformation_tasks = [task for task in task_set
                             if task['task_label']=='elastic deformation']
        opt_struct_tasks.reverse()
        if opt_struct_tasks and deformation_tasks:
            tasks_by_opt_task.append((opt_struct_tasks[-1], deformation_tasks))
        else:
            logger.warning("No structure opt matching tasks")
    return tasks_by_opt_task


def group_by_parent_lattice(docs, tol=1e-6):
    """
    Groups a set of documents by parent lattice equivalence

    Args:
        docs ([{}]): list of documents e. g. dictionaries or cursor
        tol (float): tolerance for equivalent lattice finding using,
            np.allclose, default 1e-10
    """
    docs_by_lattice = []
    for doc in docs:
        sim_lattice = get_mongolike(doc, "output.structure.lattice.matrix")

        if "deformation" in doc['task_label']:
            # Note that this assumes only one transformation, deformstructuretransformation
            defo = doc['transmuter']['transformation_params'][0]['deformation']
            parent_lattice = np.dot(sim_lattice, np.transpose(np.linalg.inv(defo)))
        else:
            parent_lattice = np.array(sim_lattice)
        match = False
        for unique_lattice, lattice_docs in docs_by_lattice:
            match = np.allclose(unique_lattice, parent_lattice, atol=tol)
            if match:
                lattice_docs.append(doc)
                break
        if not match:
            docs_by_lattice.append([parent_lattice, [doc]])
    return docs_by_lattice


def legacy_fit(strains, stresses):
    """
    Legacy fitting method for mpworks documents, intended to be temporary

    Args:
        strains: strains
        stresses: stresses

    Returns:
        elastic tensor fit using the legacy functionality

    """
    strains = [s.zeroed(0.002) for s in strains]
    return ElasticTensor.from_independent_strains(strains, stresses)


def calculate_deformation(undeformed_structure, deformed_structure):
    """

    Args:
        undeformed_structure (Structure): undeformed structure
        deformed_structure (Structure): deformed structure

    Returns:
        deformation matrix
    """
    ulatt = undeformed_structure.lattice.matrix
    dlatt = deformed_structure.lattice.matrix
    return np.transpose(np.dot(np.linalg.inv(ulatt), dlatt))


def get_warnings(elastic_tensor, structure):
    """
    Generates all warnings that apply to a fitted elastic tensor

    Args:
        elastic_tensor (ElasticTensor): elastic tensor for which
            to determine warnings
        structure (Structure): structure for which elastic tensor
            is determined

    Returns:
        list of warnings

    """
    warnings = []
    if any([s.is_rare_earth_metal for s in structure.species]):
        warnings.append("Contains a rare earth element")
    eigs, eigvecs = np.linalg.eig(elastic_tensor.voigt)
    if np.any(eigs < 0.0):
        warnings.append("Elastic tensor has a negative eigenvalue")
    c11, c12, c13 = elastic_tensor.voigt[0, 0:3]
    c23 = elastic_tensor.voigt[1, 2]

    # TODO: these should be revisited at some point, are they complete?
    #       I think they might only apply to cubic systems
    if not (abs((c11 - c12) / c11) < 0.05 or c11 < c12):
        warnings.append("c11 and c12 are within 5% or c12 is greater than c11")
    if not (abs((c11 - c13) / c11) < 0.05 or c11 < c13):
        warnings.append("c11 and c13 are within 5% or c13 is greater than c11")
    if not (abs((c11 - c23) / c11) < 0.05 or c11 < c23):
        warnings.append("c11 and c23 are within 5% or c23 is greater than c11")

    moduli = ["k_voigt", "k_reuss", "k_vrh", "g_voigt", "g_reuss", "g_vrh"]
    moduli_array = np.array([getattr(elastic_tensor, m) for m in moduli])
    if np.any(moduli_array) < 2:
        warnings.append("One or more K, G below 2 GPa")

    return warnings


def generate_formula_dict(materials_store, query=None):
    """
    Function that generates a nested dictionary of structures
    keyed first by formula and then by material_id using
    mongo aggregation pipelines

    Args:
        materials_store (Store): store of materials

    Returns:
        Nested dictionary keyed by formula-mp_id with structure values.

    """
    props = ["pretty_formula", "structure", "material_id"]
    results = list(materials_store.groupby("pretty_formula", properties=props,
                                           criteria=query))
    formula_dict = {}
    for result in tqdm.tqdm(results):
        formula = result['_id']['pretty_formula']
        material_ids = [d['material_id'] for d in result['docs']]
        structures = [d['structure'] for d in result['docs']]

        formula_dict[formula] = dict(zip(material_ids, structures))
    return formula_dict



from maggma.builder import Builder
from pydash import py_

__author__ = "Shyam Dwaraknath"
__email__ = "shyamd@lbl.gov"


class TaskTagger(Builder):

    def __init__(self, tasks, task_types, **kwargs):
        """
        Creates task_types from tasks and type definitions

        Args:
            tasks (Store): Store of task documents
            task_types (Store): Store of task_types for tasks
        """
        self.tasks = tasks
        self.task_types = task_types

        super().__init__(sources=[tasks], targets=[task_types], **kwargs)

    def get_items(self):
        """
        Returns all task docs and tag definitions to process

        Returns:
            generator or list of task docs and tag definitions
        """

        # Determine tasks to process.
        self.logger.info("Determining tasks to process")
        all_task_ids = self.tasks.distinct("task_id", {"state": "successful"})
        previous_task_ids = self.task_types.distinct("task_id")
        to_process = list(set(all_task_ids) - set(previous_task_ids))

        self.logger.info("Yielding task documents")
        for task_id in to_process:
            yield self.tasks.query_one(criteria={"task_id": task_id}, properties=["task_id", "orig_inputs"])

    def process_item(self, item):
        """
        Find the task_type for the item

        Args:
            item (dict): a (projection of a) task doc
        """
        tt = task_type(item["orig_inputs"])
        return {"task_id": item["task_id"], "task_type": tt}

    def update_targets(self, items):
        """
        Inserts the new task_types into the task_types collection

        Args:
            items ([dict]): task_type dicts to insert into task_types collection
        """
        with_task_type, without_task_type = py_.partition(items, lambda i: i["task_type"])
        if without_task_type:
            self.logger.error("No task type found for {}".format(without_task_type))
        if len(with_task_type) > 0:
            self.task_types.update(with_task_type)


def task_type(inputs, include_calc_type=True):
    """
    Determines the task_type

    Args:
        inputs (dict): inputs dict with an incar, kpoints, potcar, and poscar dictionaries
        include_calc_type (bool): whether to include calculation type
            in task_type such as HSE, GGA, SCAN, etc.
    """

    calc_type = ""

    incar = inputs.get("incar", {})

    if include_calc_type:
        if incar.get("LHFCALC", False):
            calc_type += "HSE "
        elif incar.get("METAGGA", "") == "SCAN":
            calc_type += "SCAN "
        elif incar.get("LDAU", False):
            calc_type += "GGA+U "
        else:
            calc_type += "GGA "

    if incar.get("ICHARG", 0) > 10:
        if len(list(filter(None, inputs.get("kpoints", {}).get("labels", [])))) > 0:
            return calc_type + "NSCF Line"
        else:
            return calc_type + "NSCF Uniform"

    if incar.get("LEPSILON", False):
        return calc_type + "Static Dielectric"

    if incar.get("NSW", 1) == 0:
        return calc_type + "Static"

    if incar.get("ISIF", 2) == 3 and incar.get("IBRION", 0) > 0:
        return calc_type + "Structure Optimization"

    if incar.get("ISIF", 3) == 2 and incar.get("IBRION", 0) > 0:
        return calc_type + "Deformation"

    return ""



import logging
from datetime import datetime
from itertools import chain, combinations

from pymatgen import Structure, Composition
from pymatgen.entries.compatibility import MaterialsProjectCompatibility
from pymatgen.entries.computed_entries import ComputedEntry
from pymatgen.analysis.phase_diagram import PhaseDiagram, PhaseDiagramError
from pymatgen.analysis.structure_analyzer import oxide_type, sulfide_type

from maggma.builder import Builder

__author__ = "Shyam Dwaraknath <shyamd@lbl.gov>"


class ThermoBuilder(Builder):
    def __init__(self, materials, thermo, query=None, compatibility=MaterialsProjectCompatibility("Advanced"),
                 **kwargs):
        """
        Calculates thermodynamic quantities for materials from phase diagram constructions

        Args:
            materials (Store): Store of materials documents
            thermo (Store): Store of thermodynamic data such as formation energy and decomposition pathway
            query (dict): dictionary to limit materials to be analyzed
            compatibility (PymatgenCompatability): Compatability module to ensure energies are compatible
        """

        self.materials = materials
        self.thermo = thermo
        self.query = query if query else {}
        self.compatibility = compatibility
        self.completed_tasks = set()
        super().__init__(sources=[materials], targets=[thermo], **kwargs)

    def get_items(self):
        """
        Gets sets of entries from chemical systems that need to be processed 

        Returns:
            generator of relevant entries from one chemical system
        """
        self.logger.info("Thermo Builder Started")

        # All relevant materials that have been updated since thermo props were
        # last calculated
        q = dict(self.query)
        q.update(self.materials.lu_filter(self.thermo))
        q.update({"chemsys": {"$exists": 1}})

        comps = self.materials.distinct("chemsys", q)

        self.logger.info("Found {} compositions with new/updated materials".format(len(comps)))

        # Only yields maximal super sets: e.g. if ["A","B"] and ["A"] are both in the list, will only yield ["A","B"]
        # as this will calculate thermo props for all ["A"] compounds
        processed = set()

        # Start with the largest set to ensure we don"t miss superset/subset
        # relations
        for chemsys in sorted(comps, key=lambda x: len(x.split("-")), reverse=True):
            if chemsys not in processed:
                processed |= chemsys_permutations(chemsys)
                yield self.get_entries(chemsys)

    def get_entries(self, chemsys):
        """
        Get all entries in a chemsys from materials

        Args:
            chemsys(str): a chemical system represented by string elements seperated by a dash (-)

        Returns:
            set(ComputedEntry): a set of entries for this system
        """

        self.logger.info("Getting entries for: {}".format(chemsys))

        new_q = dict(self.query)
        new_q["chemsys"] = {"$in": list(chemsys_permutations(chemsys))}
        fields = ["structure", self.materials.key, "thermo.energy", "unit_cell_formula", "calc_settings"]
        data = list(self.materials.query(fields, new_q))

        all_entries = []

        for d in data:
            entry = ComputedEntry(
                Composition(d["unit_cell_formula"]),
                d["thermo"]["energy"],
                0.0,
                parameters=d["calc_settings"],
                entry_id=d[self.materials.key],
                data={
                    "oxide_type": oxide_type(Structure.from_dict(d["structure"]))
                })

            all_entries.append(entry)

        self.logger.info("Total entries in {} : {}".format(chemsys, len(all_entries)))

        return all_entries

    def process_item(self, item):
        """
        Process the list of entries into a phase diagram

        Args:
            item (set(entry)): a list of entries to process into a phase diagram

        Returns:
            [dict]: a list of thermo dictionaries to update thermo with
        """
        entries = self.compatibility.process_entries(item)

        try:
            pd = PhaseDiagram(entries)

            docs = []

            for e in entries:
                (decomp, ehull) = \
                    pd.get_decomp_and_e_above_hull(e)

                d = {
                    self.thermo.key: e.entry_id,
                    "thermo": {
                        "formation_energy_per_atom": pd.get_form_energy_per_atom(e),
                        "e_above_hull": ehull,
                        "is_stable": e in pd.stable_entries
                    }
                }

                if d["thermo"]["is_stable"]:

                    d["thermo"]["eq_reaction_e"] = pd.get_equilibrium_reaction_energy(e)
                else:
                    d["thermo"]["decomposes_to"] = [{
                        "material_id": de.entry_id,
                        "formula": de.composition.formula,
                        "amount": amt
                    } for de, amt in decomp.items()]
                d["thermo"]["entry"] = e.as_dict()
                d["thermo"]["explanation"] = self.compatibility.get_explanation_dict(e)
                docs.append(d)
        except PhaseDiagramError as p:
            print(e.as_dict())
            self.logger.warning("Phase diagram error: {}".format(p))
            return []

        return docs

    def update_targets(self, items):
        """
        Inserts the thermo docs into the thermo collection

        Args:
            items ([[dict]]): a list of list of thermo dictionaries to update
        """
        items = list(filter(None, chain.from_iterable(items)))  # flatten out lists
        items = list({v[self.thermo.key]: v for v in items}.values())  # check for duplicates within this set
        items = [i for i in items if i[self.thermo.key] not in self.completed_tasks]  # Check if already updated this run

        self.completed_tasks |= {i[self.thermo.key] for i in items}

        if len(items) > 0:
            self.logger.info("Updating {} thermo documents".format(len(items)))
            self.thermo.update(docs=items)
        else:
            self.logger.info("No items to update")


def chemsys_permutations(chemsys):
    # Fancy way of getting every unique permutation of elements for all
    # possible number of elements:
    elements = chemsys.split("-")
    return {"-".join(sorted(c)) for c in chain(*[combinations(elements, i) for i in range(1, len(elements) + 1)])}



import os
import glob

from pymatgen.analysis.local_env import *
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.command_line.critic2_caller import Critic2Caller
from pymatgen.command_line.bader_caller import bader_analysis_from_path
from monty.os.path import which
from maggma.builder import Builder
from maggma.validator import StandardValidator

from monty.json import jsanitize

from emmet.vasp.task_tagger import task_type

__author__ = "Matthew Horton <mkhorton@lbl.gov>"


class TopologyValidator(StandardValidator):

    @property
    def schema(self):
        return {
            "type": "object",
            "properties":
                {
                    "task_id": {"type": "string"},
                    "method": {"type": "string"},
                    "successful": {"type": "boolean"}
                },
            "required": ["task_id", "method", "successful"]
        }

    @property
    def msonable_keypaths(self):
        return {"graph": StructureGraph}


class TopologyBuilder(Builder):
    def __init__(self, tasks, topology, bader,
                 query=None, use_chgcars=True,
                 critic2_settings=None,
                 **kwargs):
        """
        Builder to perform topological analysis of materials, either using
        the Bader method or through local environment nearest-neighbor
        strategies.

        This only requires tasks store to be able to look up
        original path and stored volumetric data. If use_chgcars is
        False, it won't be read.

        Topology store contains bonding information using various
        strategies, including those in pymatgen.analysis.local_env
        and using the critic2 tool.

        A topological analysis using Henkelman's bader tool is also
        done when use_chgcars=True, but since bonding information is
        not provided, the output format is different and so bader
        output is stored in a separate Store.

        Args:
            tasks (Store): Store of tasks documents
            topology (Store): Store of topology data
            bader (Store): Store of bader data
            query (dict): dictionary to limit materials to be analyzed
            use_chgcars (bool): if True, will look for charge densities
            critic2_settings (dict): user_input_settings for critic2
            in task directories and run bader and critic2 if found
        """

        self.tasks = tasks
        self.topology = topology
        self.bader = bader
        self.query = query if query else {}

        topology.validator = TopologyValidator()

        self.use_chgcars = use_chgcars
        self.critic2_settings = critic2_settings

        self.bader_available = which('bader')
        self.critic2_available = which('critic2')

        super().__init__(sources=[tasks],
                         targets=[topology, bader],
                         **kwargs)

        if not use_chgcars:
            self.logger.info("Not performing analysis of charge densities.")

        if use_chgcars and not self.bader_available:
            self.logger.error("bader binary not found! "
                                "TopologyBuilder will not be able to perform a full analysis.")
            self.logger.error("If running for Materials Project, try adding "
                                "/project/projectdirs/matgen/emmet_bin/cori to your path.")

        if not self.critic2_available:
            self.logger.error("critic2 binary not found! "
                                "TopologyBuilder will not be able to perform a full analysis.")
            self.logger.error("If running for Materials Project, try adding "
                                "/project/projectdirs/matgen/emmet_bin/cori to your path.")

    def get_items(self):
        """
        Gets all materials that need topology analysis
        """

        self.logger.info("Topology Builder Started")

        # All relevant materials that have been updated since topology
        # was last calculated
        q = dict(self.query)
        q.update(self.tasks.lu_filter(self.topology))
        tasks = self.tasks.query(criteria=q,
                                 properties=["task_id", "input.incar",
                                             "output.structure", "calcs_reversed"])

        self.logger.info("Found {} new tasks for topological analysis".format(tasks.count()))

        for task in tasks:
            if "Static" in task_type(task['input']['incar']):
                yield task

    @staticmethod
    def rewrite_dir(original_dir):
        """
        Stub. If locations of original folders has been moved,
        this function will help locate them. In practice, this
        will likely mean extracting the name of the launch dir,
        and finding this on a new host.

        Args:
            original_dir: original directory in task doc

        Returns: rewritten directory
        """

        # remove hostname if it's present, assumes builder runs on same host
        # or has access to the root_dir
        new_dir = original_dir.split(':')[1] if ':' in original_dir else original_dir

        return new_dir

    def process_item(self, item):
        """
        Calculates StructureGraphs (bonding information) for a material
        """

        topology_docs = []
        bader_doc = None

        task_id = item['task_id']
        structure = Structure.from_dict(item['output']['structure'])

        # bonding first

        self.logger.debug("Calculating bonding for {}".format(task_id))

        # try all local_env strategies
        strategies = NearNeighbors.__subclasses__()
        for strategy in strategies:
            try:
                topology_docs.append({
                    'task_id': task_id,
                    'method': strategy.__name__,
                    'graph': StructureGraph.with_local_env_strategy(structure,
                                                                    strategy()).as_dict(),
                    'successful': True
                })
            except Exception as e:

                topology_docs.append({
                    'task_id': task_id,
                    'method': strategy.__name__,
                    'successful': False,
                    'error_message': str(e)
                })

                self.logger.warning(e)
                self.logger.warning("Failed to calculate bonding for {} using "
                                      "{} local_env strategy.".format(task_id,
                                                                      strategy.__name__))

        # and also critic2 with sum of atomic charge densities
        if self.critic2_available:

            try:
                c2 = Critic2Caller(structure,
                                   user_input_settings=self.critic2_settings)

                topology_docs.append({
                    'task_id': task_id,
                    'method': 'critic2_promol',
                    'graph': c2.output.structure_graph().as_dict(),
                    'succesful': True
                })

            except Exception as e:

                topology_docs.append({
                    'task_id': task_id,
                    'method': 'critic2_promol',
                    'critic2_settings': self.critic2_settings,
                    'successful': False,
                    'error_message': str(e)
                })

                self.logger.warning(e)
                self.logger.warning("Failed to calculate bonding for {} using "
                                      "critic2 and sum of atomic charge densities.".format(task_id))

        if self.use_chgcars:

            root_dir = item['calcs_reversed'][0]['dir_name']
            root_dir = self.rewrite_dir(root_dir)

            if not os.path.isdir(root_dir):

                self.logger.error("Cannot find or cannot access {} for {}."
                                    .format(root_dir, task_id))

            else:

                if 'output_file_paths' in item['calcs_reversed'][0]:
                    # we know what output files we have

                    paths = item['calcs_reversed'][0]['output_file_paths']
                    chgcar = paths.get('chgcar', None)
                    aeccar0 = paths.get('aeccar0', None)
                    aeccar2 = paths.get('aeccar2', None)

                else:
                    # we have to search manually

                    self.logger.info("Searching {} for charge density files for {}."
                                       .format(root_dir, task_id))

                    chgcar = glob.glob(root_dir+'/*CHGCAR*')
                    chgcar = chgcar[0] if chgcar else None
                    aeccar0 = glob.glob(root_dir+'/*AECCAR0*')
                    aeccar0 = aeccar0[0] if aeccar0 else None
                    aeccar2 = glob.glob(root_dir+'/*AECCAR2*')
                    aeccar2 = aeccar2[0] if aeccar2 else None

                if chgcar and aeccar0 and aeccar2:

                    try:
                        c2 = Critic2Caller(structure,
                                           user_input_settings=self.critic2_settings)

                        topology_docs.append({
                            'task_id': task_id,
                            'method': 'critic2_chgcar',
                            'graph': c2.output.structure_graph().as_dict(),
                            'critic2_settings': self.critic2_settings,
                            'critic2_stdout': c2._stdout,
                            'successful': True
                        })

                    except Exception as e:

                        topology_docs.append({
                            'task_id': task_id,
                            'method': 'critic2_chgcar',
                            'critic2_settings': self.critic2_settings,
                            'successful': False,
                            'error_message': str(e)
                        })

                        self.logger.warning(e)
                        self.logger.warning("Failed to calculate bonding for {} "
                                              "using critic2 from CHGCAR.".format(task_id))

                    if self.bader_available:

                        try:
                            bader_doc = bader_analysis_from_path(root_dir)
                            bader_doc['task_id'] = task_id
                            bader_doc['successful'] = True
                        except Exception as e:

                            bader_doc = {
                                'task_id': task_id,
                                'successful': False,
                                'error_message': str(e)
                            }

                            self.logger.warning(e)
                            self.logger.warning("Failed to perform bader analysis "
                                                  "for {}".format(task_id))


                else:
                    self.logger.warning("Not all files necessary for charge analysis "
                                          "are present.")
                    if not chgcar:
                        self.logger.warning("Could not find CHGCAR for {}.".format(task_id))
                    else:
                        self.logger.warning("CHGCAR found for {}, but AECCAR0 "
                                              "or AECCAR2 not present.".format(task_id))

        return {
            'topology_docs': topology_docs,
            'bader_doc': bader_doc
        }

    def update_targets(self, items):

        self.logger.info("Updating topology documents")
        items = jsanitize(items)

        for item in items:

            self.topology.update(item['topology_docs'], key=['task_id', 'method'])
            if item['bader_doc']:
                self.bader.update(item['bader_doc'], key=['task_id', 'method'])


import io

from bson.binary import Binary
from maggma.builder import Builder
from pydash import py_

import matplotlib
from monty.json import MontyDecoder
from pymatgen.phonon.plotter import PhononBSPlotter

matplotlib.use('agg')


class PhononDispersionPlotter(Builder):
    def __init__(self, pmg_docs, web_docs, images,
                 add_filter=None, ignore_lu=False, **kwargs):
        """
        Produce docs for interactive and static plots of phonon dispersion.

        Args:
            pmg_docs (Store): source of serialized
                `pymatgen.phonon.bandstructure.PhononBandStructure` objects.
            web_docs (Store): target for data needed by interactive plots.
            images (Store): target for png images of phonon dispersion plots.
            add_filter (dict): MongoDB filter to add to default last-updated
                filter.
            ignore_lu (bool): Ignore pmg_docs.lu_filter when getting items.
                Useful for forcing rebuilds given `add_filter`.
        """
        self.pmg_docs = pmg_docs
        self.web_docs = web_docs
        self.images = images
        self.add_filter = add_filter if add_filter else {}
        self.ignore_lu = ignore_lu
        super().__init__(
            sources=[pmg_docs], targets=[web_docs, images], **kwargs)

    def get_items(self):
        lu_filter = self.pmg_docs.lu_filter([self.web_docs, self.images])
        filter_ = {} if self.ignore_lu else lu_filter.copy()
        filter_.update(self.add_filter)
        self.logger.info("Filtering pmg_docs by {}".format(filter_))
        cursor = self.pmg_docs.query(criteria=filter_)
        self.logger.info("Found {} pmg_docs to process".format(cursor.count()))
        if cursor.count() == 0:
            n_updated = self.pmg_docs.query(criteria=lu_filter).count()
            self.logger.debug("{} updated pmg_docs that do not match"
                              " `add_filter`".format(n_updated))
        return cursor

    def process_item(self, item):
        mp_id = item['mp-id']
        self.logger.debug("Processing {}".format(mp_id))
        decoder = MontyDecoder()
        ph_bs = decoder.process_decoded(item['ph_bs'])
        web_doc = ph_bs.as_phononwebsite()
        plotter = PhononBSPlotter(ph_bs)
        ylim = (0, max(py_.flatten_deep(plotter.bs_plot_data()['frequency'])))
        filelike = io.BytesIO()
        plotter.save_plot(filelike, ylim=ylim, img_format="png")
        image = Binary(filelike.getvalue())
        filelike.close()
        return dict(mp_id=mp_id, web_doc=web_doc, image=image)

    def update_targets(self, items):
        self.web_docs.ensure_index("mp-id", unique=True)
        self.images.ensure_index("mp-id", unique=True)
        web_docs = [{"mp-id": item["mp_id"], "ph_bs": item["web_doc"]}
                    for item in items]
        self.web_docs.update(web_docs)
        images = [{"mp-id": item["mp_id"], "plot": item["image"]}
                  for item in items]
        self.images.update(images)
        mp_ids = py_.pluck(items, "mp_id")
        self.logger.info("Updated targets for {}".format(mp_ids))

    def validate_targets(self, n=0):
        self.logger.info("Validating {}".format(
            "all" if not n else "with sample size {}".format(n)))
        ids = self.web_docs.distinct("mp-id")
        sample_ids = py_.sample_size(ids, n) if n else ids
        criteria = {"mp-id": {"$in": sample_ids}}
        web_docs = self.web_docs.query(criteria=criteria)
        n_web_docs = web_docs.count()
        assert n_web_docs == len(sample_ids), "mp-id not unique in web_docs"
        images =self.images.query(criteria=criteria)
        n_images = images.count()
        assert n_web_docs == n_images, "counts for images and web_docs differ"
        for doc in web_docs:
            assert "ph_bs" in doc and doc["ph_bs"] is not None
        self.logger.info("Validated {} of web_docs".format(n_web_docs))
        for doc in images:
            assert "plot" in doc and doc["plot"] is not None
        self.logger.info("Validated {} of images".format(n_images))
